//
// Function to build adjacency lists for triangulated data structures
//

//  The data types (EDGE and FACET) will need to be modified
//  to fit whatever framework you are using.


typedef struct tagEDGE {
 long Vid[2];     // identity of vertices to which the edge is attached
 long Fid[2];     // list to identify the facets on either side of edge   *
 //
 // add other members as required
} EDGE;

typedef struct tagFACET {
 long Vid[3];     // identity of vertices to which facet is attached
 long Fid[3];     // list to indentity facets with an edge in common      *
 long Eid[3];     // list of the identity od edges surrounding the facet  *
 //
 // add other members as required
} FACET;

// The items labelled * above are filled in by the algorithm. If there are
// no adjacent items -1 is recorded. Identifiers are integers >= 0

///////// The following data type is used internally by the function
///////// to record temporary information

typedef struct tagTEMPV {
 long *Fadj;      // list of facets adjacent to this vertex
 long nF;         // number of facets adjacent to this vertex
 long *Eadj;      // list of edges adjacent to this vertex
 long nE;         // number of edges adjacent to this vertex
} TEMPV;

///////// The following are standard functions and definitions that
///////// are usually provided as part of a compilers library.
///////// Substitute then with the ones you wish to use.

#define NULL 0
extern void *malloc(long X);      // allocate X bytes, return pointer to it.
// update memory block pointed to by P so that it has X bytes.
extern void *realloc(void *P, long X);
extern void free(void *P);        // release memory block pointed to by P

// Function to fill out the adjacency information in the EDGE and FACET
// structures. Algorithms for Delaunay triangulation and Boolean Modelling
// will require to call upon it.

void MakeAdjacencyData(    // Parameters
  long  nVertex,           // Number of vertices
  long  nEdges,            // Number of edges
  EDGE  *EdgeList,         // List of edge structures
  long  nFacets,           // Number of facets
  FACET *FacetList)        // List of facet structures
{
 TEMPV *vlist=NULL;
 long i,j,k,l,n,i0,i1,t,u,*list;
 //
 // Initialize the temporary data structure
 //
 if((vlist=(TEMPV *)malloc(nVertex*sizeof(TEMPV))) == NULL)return;
 for(i=0;i<nVertex;i++)
 {
   vlist[i].nF=0;
   vlist[i].nE=0;
   vlist[i].Fadj=NULL;
   vlist[i].Eadj=NULL;
 }
 //
 // Make the list of facets adjacent to each vertex.
 //
 for(i=0;i<nFacets;i++)
 {
   for(j=0;j<3;j++)  // for all vertices round facet
   {
     k=FacetList[i].Vid[j]; // vertices of facet i
     list=vlist[k].Fadj;
     n=vlist[k].nF;
     // extend the list
     if(list == NULL)
     {
       if((list=(long *)malloc(sizeof(long))) == NULL)return;
     }
     else
     {
       if((list=(long *)realloc(list,(n+1)*sizeof(long))) == NULL)return;
     }
     //
     // add the identity of facet i to list for vertex  k
     //
     list[n]=i;
     vlist[k].Fadj=list;
     vlist[k].nF=n+1;
   }
 }
 //
 // Make the list of edges adjacent to each vertex.
 //
 for(i=0;i<nEdges;i++)
 {
   for(j=0;j<2;j++)
   {
     k=EdgeList[i].Vid[j];
     list=vlist[k].Eadj;
     n=vlist[k].nE;
     if(list == NULL)
     {
       if((list=(long *)malloc(sizeof(long))) == NULL)return;
     }
     else {
       if((list=(long *)realloc(list,(n+1)*sizeof(long))) == NULL)return;
     }
     list[n]=i;
     vlist[k].Eadj=list;
     vlist[k].nE=n+1;
   }
 }
 //
 // Now build the list of facets adjacent to each facet
 //
 for(i=0;i<nFacets;i++)
 {
   for(l=0;l<3;l++)   // for each edge of facet i
   {
     FacetList[i].Fid[l] = -1;  // default that no facet is adjacent
     if(l == 0)
     {
       i0=FacetList[i].Vid[0];  i1=FacetList[i].Vid[1];
     }
     else if(l == 1)
     {
       i0=FacetList[i].Vid[1];  i1=FacetList[i].Vid[2];
     }
     else if(l == 2)
     {
       i0=FacetList[i].Vid[2];  i1=FacetList[i].Vid[0];
     }
     if(vlist[i0].nF > 0 && vlist[i1].nF > 0) // some facets attached
     {
       //
       // check all possible combinations
       //
       for(j=0;j<vlist[i0].nF;j++)
       {
         t=vlist[i0].Fadj[j];
         for(k=0;k<vlist[i1].nF;k++)
         {
           u=vlist[i1].Fadj[k];
           if(u == t && u != i) // this is the one  vertices i1 and i2 common
           {
             FacetList[i].Fid[l]=u;  // record facet identity
             goto JUMP_POINT1;
           }
         }
       }
     }
     JUMP_POINT1:;
   }
 }
 //
 // Now build the list of facets adjacent to each edge - its basically
 // the same as above
 //
 for(i=0;i<nEdges;i++)
 {
   i0=EdgeList[i].Vid[0];
   i1=EdgeList[i].Vid[1];
   n=0; // count number of adjacent facets found for edge  i
   EdgeList[i].Fid[0] = -1;  // default no faces adjacent
   EdgeList[i].Fid[1] = -1;
   if(vlist[i0].nF > 0 && vlist[i1].nF) // some possibilities
   {
     for(j=0;j<vlist[i0].nF;j++)
     {
       t=vlist[i0].Fadj[j];
       for(k=0;k<vlist[i1].nF;k++)
       {
         u=vlist[i1].Fadj[k];
         if(u == t)  // two faces with a common edge between vertices i0 i1
         {
           EdgeList[i].Fid[n]=u;  // record facet identity
           n++;
           if(n == 2)goto JUMP_POINT2;  // two adjacent facets found
         }
       }
     }
     JUMP_POINT2:;
   }
 }
 //
 // Build the list of edges that surround a facet - there will always
 // be three for a triangular facet. The argument is very similar to
 // that used to determine adjacent Facets
 //
 for(i=0;i<nFacets;i++)
 {
   for(l=0;l<3;l++) // for each edge of facet  i
   {
     if(l == 0)
     {
       i0=FacetList[i].Vid[0];  i1=FacetList[i].Vid[1];
     }
     else if(l == 1)
     {
       i0=FacetList[i].Vid[1];  i1=FacetList[i].Vid[2];
     }
     else if(l == 2)
     {
       i0=FacetList[i].Vid[2];  i1=FacetList[i].Vid[0];
     }
     FacetList[i].Eid[l] = -1; // default - this should never be needed
     if(vlist[i0].nE > 0 && vlist[i1].nE > 0)
     {
       for(j=0;j < vlist[i0].nE;j++)
       {
         t=vlist[i0].Eadj[j];
         for(k=0;k < vlist[i1].nE;k++)
         {
           u=vlist[i1].Eadj[k];
           if(t == u) // only one edge can have i0 and i1 in common
           {
             FacetList[i].Eid[l]=t;
             goto JUMP_POINT3; // no need to look any further
           }
         }
       }
     }
     JUMP_POINT3:;
   }
 }
 //
 // release the temporary resources
 //
 for(i=0;i<nVertex;i++)
 {
   if(vlist[i].Fadj != NULL)free(vlist[i].Fadj);
   if(vlist[i].Eadj != NULL)free(vlist[i].Eadj);
 }
 free(vlist);
 //
 // All done
 //
 return;
}
