//
// A very basic implementation of the Perlin noise table. Two functions
// are required. The first initializes a three dimensional table with
// random numbers. The second uses a simple linear interpolation to return
// a noise value from the table given an (x,y,z) coordinate.
//

///////// The following are standard functions and definitions that are
///////// usually part of a compilers library.
///////// Substitute then with the ones you wish to use.


// generate a ramdom integer in the range 0 to 32767
extern int rand(void);

// compute the largest integer not greater than "x"
extern float floor( float x);

////////////////////////////////////////////////////////////////////////


// The size of the noise table - make it larger for more accurate noise
#define NOISE_MAX 10

// The noise table
unsigned short TableOfNoise[NOISE_MAX+1][NOISE_MAX+1][NOISE_MAX+1];

////////////// The functions ////////////////////////


void InitializeNoiseTable(void)
{
 //
 // Initialize the noise table
 //
 long x,y,z,xx,yy,zz;
 for(x=0;x<=NOISE_MAX;x++)
 for(y=0;y<=NOISE_MAX;y++)
 for(z=0;z<=NOISE_MAX;z++)
 {
   //
   // Double randomize
   //
   TableOfNoise[x][y][z]=(unsigned short)rand();
   if(x == NOISE_MAX)xx=0;
   else              xx=x;
   if(y == NOISE_MAX)yy=0;
   else              yy=y;
   if(z == NOISE_MAX)zz=0;
   else              zz=z;
   TableOfNoise[x][y][z]=TableOfNoise[xx][yy][zz];
 }
}


float GetNoiseValue(float x, float y, float z)
{
 //
 // Interpolate to return a noise value at a non integer coordinate
 // from its nearest neighbors in the "TableOfNoise", do this in steps
 //
 long ix,iy,iz,n;
 float ox,oy,oz,n00,n01,n10,n11,n0,n1;
 //
 // Make sure x,y,z are always positive
 //
 ix=((long)x+30000)%NOISE_MAX;
 iy=((long)y+30000)%NOISE_MAX;
 iz=((long)z+30000)%NOISE_MAX;
 ox=x-floor(x); oy=y-floor(y); oz=z-floor(z);
 //
 // along x at base z base y
 //
 n=TableOfNoise[ix][iy][iz];
 n00=n+ox*(TableOfNoise[ix+1][iy][iz]-n);
 //
 // along x at upper z  base y
 //
 n=TableOfNoise[ix][iy][iz+1];
 n01=n+ox*(TableOfNoise[ix+1][iy][iz+1]-n);
 //
 // along x at base z upper y
 //
 n=TableOfNoise[ix][iy+1][iz];
 n10=n+ox*(TableOfNoise[ix+1][iy+1][iz]-n);
 //
 // along x at upper z upper y
 //
 n=TableOfNoise[ix][iy+1][iz+1];
 n11=n+ox*(TableOfNoise[ix+1][iy+1][iz+1]-n);
 //
 // at x interpolant along y at base z
 //
 n0=n00+oy*(n10-n00);
 //
 // at x interpolant along y at upper z
 //
 n1=n01+oy*(n11-n01);
 //
 // at x and y interpolants interpolate in z
 // scale the result to the interval (0 - 1)
 //
 return (n0+oz*(n1-n0))*3.052e-5;
}
