// Functions to draw basic elements in a frame buffer

#include "chap8.h"

///////// The following are standard functions that are usually part
///////// of a compilers library.
///////// Substitute then with the ones you wish to use.


extern void *malloc(long X);      // allocate X bytes, return pointer to it.
extern void free(void *P);        // release memory block pointed to by P
extern float sqrt(float D);       // return floating point square root of  D
extern long labs(long a);         // return absolute value
//
// Modulo functions
//
double fmod(double, double); // floating point modulo



//////////////////////////////////////////////////////
// Functions used to draw in a framebuffer ///////////
//////////////////////////////////////////////////////


void DrawPixelInBuffer(lpPIXEL Buffer,
                  int Width,        // Buffer dimensions
                  int Height,       //
                  int x, int y,     // pixel coordinates
                  unsigned char v,  // Brightness [0 - 255]
                  double rr,        //  Fractional components
                  double gg,        //  of the color
                  double bb)        //
{
 //
 // Combine the pixel with the pixel already in the buffer at (x,y)
 //
 int r,g,b;
 lpPIXEL S;
 //
 // Check dimensions
 //
 if(x < 0 || y < 0 || x >= Width || y >= Height)return;
 S=(Buffer+y*Width+x);
 r=(int)S->Red  +(int)((double)v*rr);
 g=(int)S->Green+(int)((double)v*gg);
 b=(int)S->Blue +(int)((double)v*bb);
 S->Red  =(r>255) ? 255:r;
 S->Green=(g>255) ? 255:g;
 S->Blue =(b>255) ? 255:b;
}

void DrawAntiAliasedPoint(
  float x,       // horizontal position of point to set
  float y,       // vertical position of point.  Row 0 is at top.
  float w,       // width and height of point.  w=1.0 => 1 screen pixel
                 // if w < 1.0 the brightness is attenuated to
                 // simulate smaller points
  int Brightness,// brightness of the pixel an integer in range [0-255]
  float Red,     // Red, Green, and Blue define the color of the pixel
  float Green,   // by giving the proportion of that primary color. Each
  float Blue,    // takes values in the range [0-1]
  long Width,    // width of image stored in "buffer"
  long Height,   // height of image stored in "buffer"
  lpPIXEL Buffer // Pointer to the framebuffer to be drawn in.
 )
{
  int xa,xb,ya,yb,i,j;
  double x2,y2,lfrac,rfrac,tfrac,bfrac,xfrac,yfrac;
  if(x < 0 || y < 0)return;
  x -= w*0.5; y -= w*0.5;
  x2 = x + w;
  y2 = y + w;
  xa = (int)x;
  xb = (int)x2;
  ya = (int)y;
  yb = (int)y2;
  lfrac = 1.0-fmod(x,1.0);
  rfrac = fmod(x2,1.0);
  tfrac = 1.0-fmod(y,1.0);
  bfrac = fmod(y2,1.0);
  if (xa==xb){
    xfrac = lfrac+rfrac-1.0;
    if (ya==yb) {
      DrawPixelInBuffer(Buffer,Width,Height,xa,ya,
                        xfrac*(tfrac+bfrac-1.0)*Brightness,Red,Green,Blue);
    }
    else {
      DrawPixelInBuffer(Buffer,Width,Height,xa,ya,
                        xfrac*tfrac*Brightness,Red,Green,Blue);
      for (j=ya+1; j<yb; j++)
        DrawPixelInBuffer(Buffer,Width,Height,xa,j,
                          xfrac*Brightness,Red,Green,Blue);
      DrawPixelInBuffer(Buffer,Width,Height,xa,yb,
                        xfrac*bfrac*Brightness,Red,Green,Blue);
    }
  }
  else {
    if (ya==yb) {
      yfrac = tfrac+bfrac-1.0;
      DrawPixelInBuffer(Buffer,Width,Height,xa,ya,
                        yfrac*lfrac*Brightness,Red,Green,Blue);
      for (i=xa+1; i<xb; i++)
        DrawPixelInBuffer(Buffer,Width,Height,i,ya,
                          yfrac*Brightness,Red,Green,Blue);
      DrawPixelInBuffer(Buffer,Width,Height,xb,ya,
                        yfrac*rfrac*Brightness,Red,Green,Blue);
    }
    else {
      DrawPixelInBuffer(Buffer,Width,Height,xa,ya,
                        tfrac*lfrac*Brightness,Red,Green,Blue);
      for (i=xa+1; i<xb; i++)
        DrawPixelInBuffer(Buffer,Width,Height,i,ya,
                          tfrac*Brightness,Red,Green,Blue);
      DrawPixelInBuffer(Buffer,Width,Height,xb,ya,
                        tfrac*rfrac*Brightness,Red,Green,Blue);
      for (j=ya+1; j<yb; j++){
        DrawPixelInBuffer(Buffer,Width,Height,xa,j,
                         lfrac*Brightness,Red,Green,Blue);
        for (i=xa+1; i<xb; i++)
          DrawPixelInBuffer(Buffer,Width,Height,i,j,
                            Brightness,Red,Green,Blue);
        DrawPixelInBuffer(Buffer,Width,Height,xb,j,
                          rfrac*Brightness,Red,Green,Blue);
      }
      DrawPixelInBuffer(Buffer,Width,Height,xa,yb,
                        bfrac*lfrac*Brightness,Red,Green,Blue);
      for (i=xa+1; i<xb; i++)
        DrawPixelInBuffer(Buffer,Width,Height,i,yb,
                          bfrac*Brightness,Red,Green,Blue);
      DrawPixelInBuffer(Buffer,Width,Height,xb,yb,
                        bfrac*rfrac*Brightness,Red,Green,Blue);
    }
  }
}

void DrawAntiAliasedLine(
  long i1,       // horizontal position (column) of start of line
  long j1,       // vertical position (row) of start of line
  long i2,       // horizontal position (column) of end of line
  long j2,       // vertical position (row) of end of line
  float w,       // width and height of pixel.  w=1.0 => 1 screen pixel
                 // if w < 1.0 the brightness is attenuated to
                 // simulate smaller Pixels
  int Bright,    // brightness of line [0 - 255]
  float Red,     // Red, Green, and Blue define the color of the line
  float Green,   // by giving the proportion of that primary color. Each
  float Blue,    // takes values in the range [0-1]
  long Width,    // with of image stored in "buffer"
  long Height,   // height of image stored in "buffer"
  lpPIXEL buffer // Pointer to the framebuffer to be drawn in.
 )
{
 long i,j,x,y,idx,idy;
 float grad,error;
 x=i1; y=j1;
 idx=i2-i1; idy=j2=j1;
 if(labs(idx) >= labs(idy))
 {
   grad=(float)idy/(float)idx; j=y;
   if(idy >= 0)error = -0.5; else error = 0.5;
   if(idx >= 0){
     for(i=x;i<x+idx;i++){
       error += grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       DrawAntiAliasedPoint((double)i,(double)j,w,
                            Bright,Red,Green,Blue,
                            Width,Height,buffer);
     }
   }
   else{
     for(i=x;i>x+idx;i--){
       error -= grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       DrawAntiAliasedPoint((double)i,(double)j,w,
                            Bright,Red,Green,Blue,
                            Width,Height,buffer);
     }
   }
 }
 else{
   grad=(double)idx/(double)idy; j=x;
   if(idx >= 0)error = -0.5; else error = 0.5;
   if(idy >= 0){
     for(i=y;i<y+idy;i++){
       error += grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       DrawAntiAliasedPoint((double)j,(double)i,w,
                            Bright,Red,Green,Blue,
                            Width,Height,buffer);
     }
   }
   else{
     for(i=y;i>y+idy;i--){
       error -= grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       DrawAntiAliasedPoint((double)j,(double)i,w,
                            Bright,Red,Green,Blue,
                            Width,Height,buffer);
     }
   }
 }
 return;
}

