//
// Simulate the appearance the flare that occurs in a camera lens
// by drawing appropriate structures in a framebuffer.
//
// 4 different models are presented
//
////////////////////////////////////////////////////////////////////

//
// Include file giving information on the framebuffer data structures
//
#include "chap8.h"

///////// Extern function to draw a point in the image buffer

extern void DrawAntiAliasedPoint(
  float x, float y,
  float w,
  int Brightness, float Red, float Green,  float Blue,
  long Width, long Height,
  lpPIXEL Buffer);

///////// The following are standard functions that are usually part
///////// of a compilers library.
///////// Substitute then with the ones you wish to use.

extern void *malloc(long X);      // allocate X bytes, return pointer to it.
extern void free(void *P);        // release memory block pointed to by P
extern float sqrt(float D);       // return floating point square root of  D
extern long labs(long a);         // return absolute value
extern float sin(float a);        // return sin of angle a
extern float cos(float a);        // return cosine of angle a
//
//  Quick copy function to copy a block of memory to another location
//
extern void memcpy(void *Destination, void *Source, long NumberOfBytes);
//
//  set a block of memory to specific value
//
extern void memset(void *Destination, int value, long NumberOfBytes);
//
// Random number functions
//
void srand(unsigned int);  // initialize the random number generator
int  rand(void);           // generate random number "C" language spec
                           // says this is in range [-32767 +32767]
//
// Defines
//
#define NULL 0

#define PI 3.141592657

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif


/////////////////////////////////
////////// Functions used locally
/////////////////////////////////

#define DIST(i,j)  sqrt((float)i * (float)i + (float)j * (float)j)

static float GetW(float l, float dx, float dy,
                   float Wmin, float Wmax, float iScale,
                   int linear)
{
 float d,f;
 d=DIST(dx,dy);
 f=d/l;
 if(linear)
 {
   if(f < 0.5)return (Wmin+2.0*f*(Wmax-Wmin))*(float)iScale;
   return (Wmin+(Wmax-Wmin)*2.0*(1.0-f))*(float)iScale;
 }
 else
 {
   float y,dx,dw;
   dw=(Wmax-Wmin);
   if(f < 0.5)f=1.0-f;
   f = 2.0*(f-0.5);
   y=Wmax-dw*f*f*f;
   return y*(float)iScale;
 }
 return 0.0;
}

static float GetC(float l, float dx, float dy,
                   float Wmin, float Wmax, float iScale)
{
 float d,f;
 d=DIST(dx,dy);
 f=d/l;
 if(f < 0.5)return Wmax;
 return (Wmin+(Wmax-Wmin)*2.0*(1.0-f));
}

static void DrawFlareInBuffer(long x, long y, long idx, long idy,
                              lpPIXEL buffer,
                              long X, long Y, float iScale,
                              float Cin,
                              float Wmin, float Wmax,
                              int linear)
{
 //
 // Draw a white "flare"  i.e. a line starting at (x,y) of lenght (dx,dy)
 // of a given intensity  "Cin" that increses in thickness from "Wmin"
 // at the ends to "Wmax" at the mid point. The increse in thickness can
 // be linear or quadratic.
 //
 long i,j;
 float grad,error,W,l,C;
 if(iScale <1.1)Cin *= 2; Cin=min(255,Cin); C=Cin;
 if(idx == 0 && idy == 0)return;
 l=DIST(idx,idy);
 if(labs(idx) >= labs(idy)){
   grad=(float)idy/(float)idx; j=y;
   if(idy >= 0)error = -0.5; else error = 0.5;
   if(idx >= 0){
     for(i=x;i<x+idx;i++){
       error += grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(float)(i-x),(float)(j-y),Wmin,Wmax,iScale,linear);
       if(!linear)C=GetC(l,(float)(i-x),(float)(j-y),Cin*0.25,Cin,1.0);
       DrawAntiAliasedPoint((float)i,(float)j,W,C,1.0,1.0,1.0,X,Y,buffer);
     }
   }
   else{
     for(i=x;i>x+idx;i--){
       error -= grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(float)(i-x),(float)(j-y),Wmin,Wmax,iScale,linear);
       if(!linear)C=GetC(l,(float)(i-x),(float)(j-y),Cin*0.25,Cin,1.0);
       DrawAntiAliasedPoint((float)i,(float)j,W,C,1.0,1.0,1.0,X,Y,buffer);
     }
   }
 }
 else{
   grad=(float)idx/(float)idy; j=x;
   if(idx >= 0)error = -0.5; else error = 0.5;
   if(idy >= 0){
     for(i=y;i<y+idy;i++){
       error += grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(float)(j-x),(float)(i-y),Wmin,Wmax,iScale,linear);
       if(!linear)C=GetC(l,(float)(j-x),(float)(i-y),Cin*0.25,Cin,1.0);
       DrawAntiAliasedPoint((float)j,(float)i,W,C,1.0,1.0,1.0,X,Y,buffer);
     }
   }
   else{
     for(i=y;i>y+idy;i--){
       error -= grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetW(l,(float)(j-x),(float)(i-y),Wmin,Wmax,iScale,linear);
       if(!linear)C=GetC(l,(float)(j-x),(float)(i-y),Cin*0.25,Cin,1.0);
       DrawAntiAliasedPoint((float)j,(float)i,W,C,1.0,1.0,1.0,X,Y,buffer);
     }
   }
 }
 return;
}

static float GetWL(float l, float dx, float dy,
                   float Wmin, float Wmax)
{
 float d,f;
 d=DIST(dx,dy);
 f=d/l;
 return (Wmax-(Wmax-Wmin)*f);
}

static void DrawFadedFlareInBuffer(long x, long y, long idx, long idy,
                              lpPIXEL buffer,
                              long X, long Y,
                              float Cin,
                              float r, float g, float b,
                              float Wmin, float Wmax,
                              int linear)
{
 //
 // Draw a flare that fades towards the end furthers from the light
 //
 long i,j;
 float grad,error,W,l,C;
 C=Cin;
 l=DIST(idx,idy);
 if(labs(idx) >= labs(idy)){
   grad=(float)idy/(float)idx; j=y;
   if(idy >= 0)error = -0.5; else error = 0.5;
   if(idx >= 0){
     for(i=x;i<x+idx;i++){
       error += grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetWL(l,(float)(i-x),(float)(j-y),Wmin,Wmax);
       DrawAntiAliasedPoint((float)i,(float)j,W,C,1.0,1.0,1.0,X,Y,buffer);
     }
   }
   else{
     for(i=x;i>x+idx;i--){
       error -= grad;
       if(idy >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetWL(l,(float)(i-x),(float)(j-y),Wmin,Wmax);
       DrawAntiAliasedPoint((float)i,(float)j,W,C,1.0,1.0,1.0,X,Y,buffer);
     }
   }
 }
 else{
   grad=(float)idx/(float)idy; j=x;
   if(idx >= 0)error = -0.5; else error = 0.5;
   if(idy >= 0){
     for(i=y;i<y+idy;i++){
       error += grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetWL(l,(float)(j-x),(float)(i-y),Wmin,Wmax);
       DrawAntiAliasedPoint((float)j,(float)i,W,C,1.0,1.0,1.0,X,Y,buffer);
     }
   }
   else{
     for(i=y;i>y+idy;i--){
       error -= grad;
       if(idx >= 0){
         if(error > 0.0){j++; error -= 1.0;}
       }
       else{
         if(error < 0.0){j--; error += 1.0;}
       }
       W=GetWL(l,(float)(j-x),(float)(i-y),Wmin,Wmax);
       DrawAntiAliasedPoint((float)j,(float)i,W,C,1.0,1.0,1.0,X,Y,buffer);
     }
   }
 }
 return;
}

static void DrawHaloInBuffer(long x, long y,
                             lpPIXEL buffer,
                             long X, long Y, long radius,
                             float maxV,
                             float R, float G, float B)
{
 //
 // mix in a halo at the screen buffer location of the light source
 //
 long i,j;
 float d;
 lpPIXEL S;
 for(i=x-radius;i<=x+radius;i++)for(j=y-radius;j<=y+radius;j++){
   if(i < 0 || i >= X || j < 0 || j >= Y)continue;
   S=(buffer + j*X + i);
   d=sqrt((float)(i-x)*(float)(i-x) + (float)(j-y)*(float)(j-y));
   if(d < radius*0.25)d=maxV;
   else d=maxV*1.3333*((float)radius-d)/(float)radius;
   d=max(0.0,d);
   S->Red   = (unsigned char)min(255.0,(float)S->Red  +d*R);
   S->Green = (unsigned char)min(255.0,(float)S->Green+d*G);
   S->Blue  = (unsigned char)min(255.0,(float)S->Blue +d*B);
 }
 return;
}

static void DrawDiskInBuffer(long x, long y,
                             lpPIXEL buffer,
                             long X, long Y, long radius,
                             float minV, float maxV,
                             float R, float G, float B){
 long i,j;
 float d;
 lpPIXEL S;
 for(i=x-radius;i<=x+radius;i++)for(j=y-radius;j<=y+radius;j++){
   if(i < 0 || i >= X || j < 0 || j >= Y)continue;
   S=(buffer + j*X + i);
   d=sqrt((float)(i-x)*(float)(i-x) + (float)(j-y)*(float)(j-y));
   if(d < (float)radius){
     d=minV+(maxV-minV)*((float)radius-d)/(float)radius;
       S->Red   = (unsigned char)min(255.0,(float)S->Red  +d*R);
     S->Green = (unsigned char)min(255.0,(float)S->Green+d*G);
     S->Blue  = (unsigned char)min(255.0,(float)S->Blue +d*B);
   }
 }
 return;
}

static void DrawDiskWithFade(long x, long y,
                             lpPIXEL buffer,
                             long X, long Y,
                             long radius, long dradius,
                             float minV, float maxV,
                             float R, float G, float B)
{
 long i,j;
 float d,dr;
 lpPIXEL S;
 for(i=x-dradius;i<=x+dradius;i++)for(j=y-dradius;j<=y+dradius;j++){
   if(i < 0 || i >= X || j < 0 || j >= Y)continue;
     S=(buffer + j*X + i);
     d=sqrt((float)(i-x)*(float)(i-x) + (float)(j-y)*(float)(j-y));
     if(d < (float)radius){
       d=minV+(maxV-minV)*((float)radius-d)/(float)radius;
       S->Red   = (unsigned char)min(255.0,(float)S->Red  +d*R);
       S->Green = (unsigned char)min(255.0,(float)S->Green+d*G);
       S->Blue  = (unsigned char)min(255.0,(float)S->Blue +d*B);
     }
     else if(d < (float)dradius){
       d -= (float)radius; dr = (float)(dradius-radius);
       d=minV*(dr-d)/dr;
       S->Red   = (unsigned char)min(255.0,(float)S->Red  +d*R);
       S->Green = (unsigned char)min(255.0,(float)S->Green+d*G);
       S->Blue  = (unsigned char)min(255.0,(float)S->Blue +d*B);
     }
 }
 return;
}

static void DrawCircleInBuffer(long x, long y,
                             lpPIXEL buffer,
                             long X, long Y,
                             long iradius, long oradius,
                             float minV, float maxV,
                             float R, float G, float B)
{
 long i,j;
 float d,dc;
 lpPIXEL S;
 dc=0.5*(float)(iradius+oradius);
 for(i=x-oradius;i<=x+oradius;i++)for(j=y-oradius;j<=y+oradius;j++){
   if(i < 0 || i >= X || j < 0 || j >= Y)continue;
   S=(buffer + j*X + i);
   d=sqrt((float)(i-x)*(float)(i-x) + (float)(j-y)*(float)(j-y));
   if(d >= (float)iradius && d <= (float)oradius){
     if(d > dc){
       d=maxV-(maxV-minV)*(d-dc)/((float)oradius-dc);
     }
     else{
       d=minV+(maxV-minV)*(d-(float)iradius)/(dc-(float)iradius);
     }
     S->Red   = (unsigned char)min(255.0,(float)S->Red  +d*R);
     S->Green = (unsigned char)min(255.0,(float)S->Green+d*G);
     S->Blue  = (unsigned char)min(255.0,(float)S->Blue +d*B);
   }
 }
 return;
}

static void DrawInterLensReflection(
            float lc, float dx, float dy, float iScale,
            long Xc, long Yc, long Width, long Height,
            lpPIXEL Buffer)
{
 //
 // Draw the series of colored disks and rings that appear like the
 // inter-lens reflections that occur as a result of lens flare.
 //
 long xBlue,yBlue,xBrown,yBrown,xGreen,yGreen,xT,yT;
 //
 // large brown disk behind the light
 //
 DrawDiskInBuffer(Xc+(long)(1.8*lc*dx),Yc+(long)(1.8*lc*dy),
                        Buffer,Width,Height,
                        (long)(120*iScale),
                        20.0,30.0,
                        1.0,0.1,0.05);
 //
 // A yellow ring
 //
 DrawCircleInBuffer(Xc+(long)(1.3*lc*dx),Yc+(long)(1.3*lc*dy),
                        Buffer,Width,Height,
                        (long)(10*iScale),(long)(13*iScale),
                        0.0,30.0,
                        1.0,1.0,0.2);
 //
 // The center of the yellow ring filled in with fainter edge
 //
 DrawDiskInBuffer(Xc+(long)(1.3*lc*dx),Yc+(long)(1.3*lc*dy),
                        Buffer,Width,Height,
                        (long)(15*iScale),
                        10.0,20.0,
                        1.0,1.0,0.2);
 //
 // three overlapping blue disks two of which are slightly offset
 //
 xBlue=Xc+(long)(0.5*lc*dx); yBlue=Yc+(long)(0.5*lc*dy);
 DrawDiskWithFade(xBlue,yBlue,
                        Buffer,Width,Height,
                        (long)(20*iScale),(long)(25*iScale),
                        40.0,60.0,
                        0.0,0.0,1.0);
 DrawDiskWithFade(xBlue+(long)(8*iScale*dx),yBlue+(long)(8*iScale*dy),
                        Buffer,Width,Height,
                        (long)(6*iScale),(long)(8*iScale),
                        40.0,44.0,
                        0.0,0.0,1.0);
 DrawDiskWithFade(xBlue-(long)(6*iScale*dx),yBlue-(long)(6*iScale*dy),
                        Buffer,Width,Height,
                        (long)(9*iScale),(long)(11*iScale),
                        40.0,44.0,
                        0.0,0.0,1.0);
 //
 // three overlapping brown disks two of which are slightly offset
 //
 xBrown=Xc-(long)(0.5*lc*dx); yBrown=Yc-(long)(0.5*lc*dy);
 DrawDiskWithFade(xBrown,yBrown,
                        Buffer,Width,Height,
                        (long)(18*iScale),(long)(21*iScale),
                        40.0,44.0,
                        1.0,0.1,0.2);
 DrawDiskWithFade(xBrown-(long)(8*iScale*dx),yBrown-(long)(8*iScale*dy),
                        Buffer,Width,Height,
                        (long)(5*iScale),(long)(7*iScale),
                        40.0,44.0,
                        1.0,0.1,0.2);
 DrawDiskWithFade(xBrown+(long)(6*iScale*dx),yBrown+(long)(6*iScale*dy),
                        Buffer,Width,Height,
                        (long)(10*iScale),(long)(12*iScale),
                        40.0,44.0,
                        1.0,0.1,0.2);
 //
 // medium sized brown disk
 //
 xBrown=Xc+(long)(0.35*lc*dx); yBrown=Yc+(long)(0.35*lc*dy);
 DrawDiskInBuffer(xBrown,yBrown,
                        Buffer,Width,Height,
                        (long)(5*iScale),
                        40.0,44.0,
                        1.0,0.1,0.2);
 //
 // a blue halo with slightly offset white and green disks
 //
 xT=Xc-(long)(0.8*lc*dx); yT=Yc-(long)(0.8*lc*dy);
 DrawHaloInBuffer(xT,yT,Buffer,Width,Height,
                        (long)(10*iScale),
                        160.0,
                        0.2,0.2,1.0);
 //
 // white halo
 //
 DrawHaloInBuffer(xT-(long)(2*iScale*dx),yT-(long)(2*iScale*dy),
                        Buffer,
                        Width,Height,
                        (long)(2*iScale),
                        64.0,
                        1.0,1.0,1.0);
 //
 // green halo
 //
 DrawDiskWithFade(xT+(long)(8*iScale*dx),yT+(long)(8*iScale*dy),
                        Buffer,Width,Height,
                        (long)(11*iScale),(long)(13*iScale),
                        40.0,44.0,
                        0.5,0.8,0.2);
 //
 // a green disk
 //
 xGreen=Xc-(long)(0.8*lc*dx); yGreen=Yc-(long)(0.8*lc*dy);
 DrawDiskInBuffer(xGreen,yGreen, // green disk
                        Buffer,Width,Height,
                        (long)(30*iScale),
                        20.0,44.0,
                        0.4,0.5,0.2);
 //
 // the outline of the green disk
 //
 DrawCircleInBuffer(xGreen,yGreen,
                        Buffer,Width,Height,
                        (long)(28*iScale),(long)(30*iScale),
                        0.0,30.0,
                        1.0,1.0,0.2);
 //
 // The rainbow effect opposite to the light
 //
 xT=Xc-(long)(1.2*lc*dx); yT=Yc-(long)(1.2*lc*dy);
 DrawCircleInBuffer(xT,yT,
                        Buffer,Width,Height,
                        (long)(60*iScale),(long)(62*iScale),
                        10.0,30.0,
                        1.0,0.2,0.2);
 DrawCircleInBuffer(xT,yT,
                        Buffer,Width,Height,
                        (long)(58*iScale),(long)(61*iScale),
                        30.0,30.0,
                        0.2,1.0,0.2);
 DrawCircleInBuffer(xT,yT,
                        Buffer,Width,Height,
                        (long)(56*iScale),(long)(59*iScale),
                        10.0,30.0,
                        0.2,0.2,1.0);
 //
 // bright dot at screen center
 //
 DrawAntiAliasedPoint((float)Width/2,(float)Height/2,
                      1.0*iScale,255,1.0,1.0,1.0,
                      Width,Height,Buffer);
 xT=Xc-(long)(0.35*lc*dx); yT=Yc-(long)(0.35*lc*dy);
 //
 // bright white off-centre halo
 //
 DrawHaloInBuffer(xT,yT,Buffer,         
                        Width,Height,
                        (long)(3*iScale),
                        255.0,
                        0.8,0.8,1.0);
}

//////////////////////////////////////////////////////////
//////////// Lens flare models ///////////////////////////
//////////////////////////////////////////////////////////



void DrawLensFlare_Model1(
  long xx,       // x coordinate of the position of the light (in the buffer)
  long yy,       // y coordinate of the position of the light (in the buffer)
  long Width,    // Width of the framebuffer
  long Height,   // height of the framebuffer
  lpPIXEL Buffer)// pointer to the framebuffer
{
 //
 // Simulate a lens flare with inter-lens reflections and random array of
 // fine spines.
 //
 int i,j,n=45;
 long Xc,Yc;
 float iScale,ascale,lscale,dlscale,d2r,lc,dx,dy;
 //
 // The basic picture is painted for an image of 320 pixels wide for wider
 // images all the features need to be scales "iScale" chages the feature
 // size.  Xc and Yc are co-ordinates at the center of the imagebuffer.
 //
 Xc=Width/2; Yc=Height/2;
 iScale=(float)Width/320;
 ascale=5.0/32767.0;
 lscale=80.0*(float)iScale;
 dlscale=80.0/32767.0*(float)iScale;
 d2r=3.1415626/180.0;
 srand(1);
 dx=(float)(xx-Xc);  dy=(float)(yy-Yc);
 lc=DIST(dx,dy);  dx /= lc; dy /= lc;
 //
 // Draw the inter-lens reflections
 //
 DrawInterLensReflection(lc,dx,dy,iScale,
                         Xc,Yc,Width,Height,Buffer);
 //
 // Draw a series of halow round the light
 //
 DrawHaloInBuffer(xx,yy,Buffer,Width,Height,
                          (long)(3*iScale),
                          255.0,
                          1.0,1.0,1.0);
 //
 // Red disk halo
 //
 DrawHaloInBuffer(xx,yy,Buffer,Width,Height,
                          (long)(20*iScale),
                          80.0,
                          1.0,0.5,0.2);
 //
 // Draw a ring halo round the light
 //
 DrawCircleInBuffer(xx,yy,
                          Buffer,Width,Height,
                          (long)(20*iScale),(long)(23*iScale),
                          50.0,80.0,
                          1.0,0.5,0.0);
 //
 // A long set of randomly directed and random length spines.
 //
 for(j=0;j<n;j++)
 {
           float a,l;
           long   idx,idy;
           a=360.0/(float)n*(float)j+ascale*(float)rand();
           a *= d2r;
           l=lscale+dlscale*(float)rand();
           idx=(long)(l*cos(a)); idy=(long)(l*sin(a));
           DrawFlareInBuffer(xx,yy,idx,idy,Buffer,Width,Height,
                             iScale,
                             20.0,     // brightness 0-255
                             0.2,0.8,  // min width max width
                             1);
 }
 //
 // Three shorter sets of randomly directed and random length spines.
 //
 for(i=0;i<3;i++)for(j=0;j<n;j++)
 {
           float a,l,C;
           long   idx,idy;
           a=360.0/(float)n*(float)j+ascale*(float)rand();
           a *= d2r;
           l=lscale+dlscale*(float)rand();
           if     (i == 0){l /= 5;  C=32.0;}
           else if(i == 1){l /= 10; C=64.0;}
           else if(i == 2){l /= 20; C=128.0;}
           idx=(long)(l*cos(a)); idy=(long)(l*sin(a));
           DrawFlareInBuffer(xx,yy,idx,idy,Buffer,Width,Height,
                             iScale,
                             C,
                             0.2,0.8,
                             1);
 }
 return;
}



void DrawLensFlare_Model2(
  long xx,       // x coordinate of the position of the light (in the buffer)
  long yy,       // y coordinate of the position of the light (in the buffer)
  long Width,    // Width of the framebuffer
  long Height,   // height of the framebuffer
  lpPIXEL Buffer)// pointer to the framebuffer
{
 //
 // Simulate a lens flare with inter-lens reflections and regularly arranged
 // star like flares emerging from the light.
 //
 int i,j,n=45;
 long Xc,Yc;
 float iScale,ascale,lscale,dlscale,d2r,lc,dx,dy;
 //
 // The basic picture is painted for an image of 320 pixels wide for wider
 // images all the features need to be scales "iScale" chages the feature
 // size.  Xc and Yc are co-ordinates at the center of the imagebuffer.
 //
 Xc=Width/2; Yc=Height/2;
 iScale=(float)Width/320;
 ascale=5.0/32767.0;
 lscale=80.0*(float)iScale;
 dlscale=80.0/32767.0*(float)iScale;
 d2r=3.1415626/180.0;
 srand(1);
 dx=(float)(xx-Xc);  dy=(float)(yy-Yc);
 lc=DIST(dx,dy);  dx /= lc; dy /= lc;
 //
 // Draw the inter-lens reflections
 //
 DrawInterLensReflection(lc,dx,dy,iScale,
                         Xc,Yc,Width,Height,Buffer);
 //
 // Draw a series of fading disks and rings round the position of the light
 //
 // White disk
 //
 DrawHaloInBuffer(xx,yy,Buffer,Width,Height,
                          (long)(20*iScale),
                          255.0,
                          1.0,1.0,1.0);
 //
 // Red disk
 //
 DrawHaloInBuffer(xx,yy,Buffer,Width,Height,
                          (long)(30*iScale),
                          80.0,
                          1.0,0.5,0.2);
 //
 // Red ring round red disk
 //
 DrawCircleInBuffer(xx,yy,
                          Buffer,Width,Height,
                          (long)(150*iScale),(long)(160*iScale),
                          50.0,80.0,
                          1.0,0.5,0.0);
 //
 // Draw a larger whitish disk round the light
 //
 DrawDiskWithFade(xx,yy,
                         Buffer,Width,Height,
                          (long)(27*iScale),
                          (long)(40*iScale),
                          100.0,100.0,
                          1.0,1.0,1.0);
 //
 // Draw a regular array of broad spines that fade near the ends and
 // increase in width towards their centers
 //
 for(j=0;j<10;j++)
 {
   //
   // Each spine is constructed by drawing overlapping simple spines
   // of increasing thickness and brightness
   //
           float a,l,ddx,ddy;
           long   xs,ys,idx,idy;
           a= -90.0 + 360.0/(float)10*(float)j;
           a *= d2r;
           l=lscale*1.5;
           ddx=l*cos(a); ddy=l*sin(a);
           idx=(long)ddx; idy=(long)ddy;
           xs=xx+(long)(ddx*0.09); ys=yy+(long)(ddy*0.09);
           DrawFlareInBuffer(xs,ys,idx,idy,Buffer,Width,Height,
                             iScale,
                             40.0,
                             0.0,1.4,
                             0);
           DrawFlareInBuffer(xs,ys,idx,idy,Buffer,Width,Height,
                             iScale,
                             20.0,
                             0.0,2.8,
                             0);
           xs=xx+(long)(ddx*0.2); ys=yy+(long)(ddy*0.2);
           DrawFlareInBuffer(xs,ys,idx,idy,Buffer,Width,Height,
                             iScale,
                             10.0,
                             0.0,4.5,
                             0);
 }
 return;
}



void DrawLensFlare_Model3(
  long xx,       // x coordinate of the position of the light (in the buffer)
  long yy,       // y coordinate of the position of the light (in the buffer)
  long Width,    // Width of the framebuffer
  long Height,   // height of the framebuffer
  lpPIXEL Buffer)// pointer to the framebuffer
{
 //
 // Simulate a les flare with inter-lens reflections and a simple halo
 // round the light.
 //
 int i,j,n=45;
 long Xc,Yc;
 float iScale,ascale,lscale,dlscale,d2r,lc,dx,dy;
 //
 // The basic picture is painted for an image of 320 pixels wide for wider
 // images all the features need to be scales "iScale" chages the feature
 // size.  Xc and Yc are co-ordinates at the center of the imagebuffer.
 //
 Xc=Width/2; Yc=Height/2;
 iScale=(float)Width/320;
 ascale=5.0/32767.0;
 lscale=80.0*(float)iScale;
 dlscale=80.0/32767.0*(float)iScale;
 d2r=3.1415626/180.0;
 srand(1);
 dx=(float)(xx-Xc);  dy=(float)(yy-Yc);
 lc=DIST(dx,dy);  dx /= lc; dy /= lc;
 //
 // Draw the inter-lens reflections
 //
 DrawInterLensReflection(lc,dx,dy,iScale,
                         Xc,Yc,Width,Height,Buffer);

 //
 // Draw a series of fading disks and rings round the position of the light
 // similar to Flare model 2
 //
 DrawHaloInBuffer(xx,yy,Buffer,Width,Height,
                          (long)(20*iScale),
                          255.0,
                          1.0,1.0,1.0);
 DrawHaloInBuffer(xx,yy,Buffer,Width,Height,
                          (long)(30*iScale),
                          80.0,
                          1.0,0.5,0.2);
 DrawCircleInBuffer(xx,yy,
                          Buffer,Width,Height,
                          (long)(150*iScale),(long)(160*iScale),
                          50.0,80.0,
                          1.0,0.5,0.0);
 return;
}



void DrawLensFlare_Model4(
  long xx,       // x coordinate of the position of the light (in the buffer)
  long yy,       // y coordinate of the position of the light (in the buffer)
  float cr,      // Red
  float cg,      // Green (0 - 1) give the color of the flare
  float cb,      // Blue
  long Width,    // Width of the framebuffer
  long Height,   // height of the framebuffer
  lpPIXEL Buffer)// pointer to the framebuffer
{
 //
 // Simulate a star filter appled to a cemera lens. This gives a
 // pronounced "star like" appearance to a source of light.
 //
 int i,j,n=45;
 long Xc,Yc;
 float iScale,ascale,lscale,dlscale,d2r,lc,dx,dy;
 //
 // The basic picture is painted for an image of 320 pixels wide for wider
 // images all the features need to be scales "iScale" chages the feature
 // size.  Xc and Yc are co-ordinates at the center of the imagebuffer.
 //
 Xc=Width/2; Yc=Height/2;
 iScale=(float)Width/320;
 ascale=5.0/32767.0;
 lscale=80.0*(float)iScale;
 dlscale=80.0/32767.0*(float)iScale;
 d2r=3.1415626/180.0;
 srand(1);
 dx=(float)(xx-Xc);  dy=(float)(yy-Yc);
 lc=DIST(dx,dy);  dx /= lc; dy /= lc;
 //
 // Draw the longer 8 points in the star
 //
 for(j=0;j<8;j++)
 {
             float a,l;
             long   idx,idy;
             a=360.0/(float)8*(float)j+22.5;
             a *= d2r;
             l=lscale;
             idx=(long)(l*cos(a)); idy=(long)(l*sin(a));
             DrawFadedFlareInBuffer(xx,yy,idx,idy,
                             Buffer,
                             Width,Height,
                             8.0,
                             cr,cg,cb,
                             4.0*iScale/2,4.5*iScale/2,
                             1);
             DrawFadedFlareInBuffer(xx,yy,idx,idy,
                             Buffer,
                             Width,Height,
                             16.0,
                             cr,cg,cb,
                             1.0*iScale/2,3.0*iScale/2,
                             1);
             DrawFadedFlareInBuffer(xx,yy,idx,idy,
                             Buffer,
                             Width,Height,
                             80.0,
                             1.0,1.0,1.0,
                             0.0,2.0*iScale/2,
                             1);
             idx /= 2; idy /=2;
             DrawFadedFlareInBuffer(xx,yy,idx,idy,
                             Buffer,
                             Width,Height,
                             16.0,
                             cr,cg,cb,
                             0.0,2.0*iScale/2,
                             1);
 }
 //
 // Draw the shorter 8 points of the star (rotated from the first)
 //
 for(j=0;j<8;j++)
 {
             float a,l;
             long   idx,idy;
             a=360.0/(float)8*(float)j;
             a *= d2r;
             l=lscale/2;
             idx=(long)(l*cos(a)); idy=(long)(l*sin(a));
             DrawFadedFlareInBuffer(xx,yy,idx,idy,
                             Buffer,
                             Width,Height,
                             64.0,
                             cr,cg,cb,
                             0.8*iScale/2,1.5*iScale/2,
                             1);
 }
 //
 // Draw a series of very short flares to "fill in" the light area.
 // and reduce the "mechanical" appearance of the flare
 //
 n=36;
 for(j=0;j<n;j++)
 {
             float a,l;
             long   idx,idy;
             a=360.0/(float)n*(float)j+ascale*(float)rand();
             a *= d2r;
             l=(lscale+dlscale*(float)rand())/30;
             idx=(long)(l*cos(a)); idy=(long)(l*sin(a));
             DrawFadedFlareInBuffer(xx,yy,idx,idy,
                             Buffer,
                             Width,Height,
                             255.0,
                             1.0,1.0,1.0,
                             0.8*iScale/2,1.2*iScale/2,
                             1);
 }
 //
 // Draw a fading disk round the light to complete the effect
 //
 DrawDiskWithFade(xx,yy,Buffer,Width,Height,
                        (long)(5*iScale), (long)(20*iScale),
                        128.0,255.0,
                        cr,cg,cb);
 return;
}

