/////////////////////////////////////////////////////////////////////////
// Dot texture following a Face Centered Cubic Pattern

BOOL FccDot_Texture(   // return TRUE if successful
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 float size = 0.5;
 float edge = 10.0;
 VECTOR c1={1.0,0.0,0.0};
 VECTOR c2={0.0,1.0,0.0};
 VECTOR c3={0.0,0.0,1.0};
 ////////////////////////////////////
  static VECTOR pp[6]={{0.0,0.5,0.5},     // location of spots relative
                       {0.5,1.0,0.5},     // to the texture cell.
                       {1.0,0.5,0.5},
                       {0.5,0.0,0.5},
                       {0.5,0.5,0.0},
                       {0.5,0.5,1.0}};
 long i;
 float r,g,b;
 float dx,dy,dz,zb,e,a;
 xx=(xx-FLOOR(xx));                      //  Relate to the texture cell
 yy=(yy-FLOOR(yy));                      //
 zz=(zz-FLOOR(zz));                      //
 zb=0.0;
 e=(edge/100.0)*size;
 r=(float)color[0];
 g=(float)color[1];
 b=(float)color[2];
 for(i=0;i<6;i++)
 {                                     //  Work through all the dots
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < size)
   {                                  // if point (xx,yy,zz) is inside
     if((size-a) > zb)                // use a "Z" buffer so that dots
     {                                // don't overlap
       zb=(size-a);
       if(i == 0 || i == 2){                // set dots on opposite cell
         r=c1[0];                           // faces so that they have the
         g=c1[1];                           // color
         b=c1[2];
       }
       else if(i == 1 || i == 3){
         r=c2[0];
         g=c2[1];
         b=c2[2];
       }
       else{
         r=c3[0];
         g=c3[1];
         b=c3[2];
       }
       if(a > e){
         mix(&r,&g,&b,(a-e)/(size-e),   // near edge mix with background
          (float)color[0],
          (float)color[1],
          (float)color[2],
          r,g,b);
       }
     }
   }
 }
 color[0]  = (unsigned char)r;      //
 color[1]  = (unsigned char)g;      //
 color[2]  = (unsigned char)b;      //
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Hexagional Close Packed Dots

BOOL HcpDot_Texture(   // return TRUE if successful
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 float size = 0.5;
 VECTOR c1={1.0,0.0,0.0};  // Red sphere
 ////////////////////////////////////
 static VECTOR pp[35]={                         // coordinates in unit cell
 /* row 1 */    {0.25,0.0,      0.0},           // of the texture spheres
                {0.75,0.0,      0.0},           // 32 are uses to ensure
                {0.0, ROOT32/2, 0.0},           // that in cases where they
                {0.5, ROOT32/2, 0.0},           // are perturbed there are
                {1.0, ROOT32/2, 0.0},           // always enought to fill
                {0.25,ROOT32,   0.0},           // the texture cell
                {0.75,ROOT32,   0.0},

 /* extras */   {-0.25,0.0,     0.0},
                { 1.25,0.0,     0.0},
                {-0.25,ROOT32,  0.0},
                { 1.25,ROOT32,  0.0},
                {-0.25,0.0,    ROOT32},
                { 1.25,0.0,    ROOT32},
                {-0.25,ROOT32, ROOT32},
                { 1.25,ROOT32, ROOT32},
                {-0.25,1.0/2.0/ROOT3,ROOT32/2.0},
                { 1.75,1.0/2.0/ROOT3,ROOT32/2.0},
                {0.0,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},
                {0.5,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},
                {1.0,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},


/* row 2 */     {0.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.25,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.75,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},

 /* row 3 */    {0.25,0.0,      ROOT32},
                {0.75,0.0,      ROOT32},
                {0.0, ROOT32/2, ROOT32},
                {0.5, ROOT32/2, ROOT32},
                {1.0, ROOT32/2, ROOT32},
                {0.25,ROOT32,   ROOT32},
                {0.75,ROOT32,   ROOT32}};
 float dx,dy,dz,zb;
 long i;
 float a,cc,r,g,b;
 VECTOR u,v;
 cc=size;
 //
 // The (u,v,w) coordinates are rescaled from a 1,1,1 unit cell to the
 // HCP cell size so that spheres look spherical and not squashed ovals.
 //
 yy /= ROOT32;
 xx=FMOD(xx,1.0);
 yy=FMOD(yy,ROOT32);
 zz /= ROOT32;
 zz=FMOD(zz,ROOT32);
 zb=0.0;
 r=color[0];
 g=color[1];
 b=color[2];
 for(i=0;i<35;i++)     // test if (xx,yy,zz) lies inside any of the spheres
 {
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if(fabs(dx) > cc)continue;
   if(fabs(dy) > cc)continue;
   if(fabs(dz) > cc)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc)
   {
     if((cc-a) > zb)
     {
       zb=(cc-a);
       r=c1[0];
       g=c1[1];
       b=c1[2];
     }
     break;
   }
 }
 color[0]  = r;
 color[1]  = g;
 color[2]  = b;
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Pimples - one sphere at center of texture cell

BOOL Pimple_Texture(   // return TRUE if successful
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 int type = 1;
 float depth = 1.0;
 float size = 0.5;
 ////////////////////////////////////
 float a;                          // all calculations are done relative
 VECTOR u,v,pp;                    // to a global frame of reference
 float rad;
 rad=size;
 VECSUM(P,0.5*U,pp)
 VECSUM(pp,0.5*V,pp)
 VECSUM(pp,0.5*W,pp)               // coord of centre of pattern
 xx=(xx-floor(xx));
 yy=(yy-floor(yy));
 zz=(zz-floor(zz));
 VECCOPY(P,v)
 VECSUM(v,xx*U,v)
 VECSUM(v,yy*V,v)
 VECSUM(v,zz*W,v)         // Coord of point on surface relative
                          // to World coordinate system.
 VECSUB(v,pp,u)           //(v-pp) is VECTOR from centre of pattern
 if((a=sqrt((u[0])*(u[0])+
            (u[1])*(u[1])+
            (u[2])*(u[2]))) < rad)
 {                        // point is inside the sphere
   float mu,moddn,rmu;
   VECTOR c,dn;
   mu=DOT(u,n);
   c[0]=pp[0]+mu*n[0];    // "c" is point on surface at center of dot
   c[1]=pp[1]+mu*n[1];
   c[2]=pp[2]+mu*n[2];
   VECSUB(v,c,dn)
   moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
   rmu=sqrt(rad*rad-mu*mu);
   if(moddn < rmu)
   {
     a=1.0-moddn/rmu;     // See text for details of calculation
     normalize(dn);
     //
     //  "a" is fraction of how close to center of dot "p" lies.
     //  Scale "a" to give different shapes to the pimples.
     //
     if(type == 0)a=sin(a*1.57079);  // rounded
     else{
       if(a < 0.5)
       {                             // studs/mesas
         if(type == 2)a=0.5;
         else if(type == 1)a=1.0-sigmoid(a,0.0,0.5);
       }
       else a=1;
     }
     a=1.0-(1.0-a)*depth;            // make pimple look more noticable
     VECSUM((1.0-a)*dn,(a)*n,n)
     normalize(n);
   }
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Dimples - this is basically the same as the Pimples - see details there.

BOOL Dimplet_Texture(  // return TRUE if successful
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 int type = 1;
 float depth = 1.0;
 float size = 0.5;
 ////////////////////////////////////
 float a;
 VECTOR u,v,pp;
 float rad;
 rad=size;
 VECSCALE(rad,u,u)
 VECSUM(P,0.5*U,pp)
 VECSUM(pp,0.5*V,pp)
 VECSUM(pp,0.5*W,pp)           // coord of centre of pattern
 xx=(xx-floor(xx));
 yy=(yy-floor(yy));
 zz=(zz-floor(zz));
 VECCOPY(P,v)
 VECSUM(v,xx*U,v)
 VECSUM(v,yy*V,v)
 VECSUM(v,zz*W,v)             // coord of point on surface (global frame)
 VECSUB(v,pp,u)               //(v-pp) VECTOR from centre of pattern
 if((a=sqrt((u[0])*(u[0])+
            (u[1])*(u[1])+
            (u[2])*(u[2]))) < rad)
 {
   float mu,moddn,rmu;
   VECTOR c,dn;
   mu=DOT(u,n);
   c[0]=pp[0]+mu*n[0];
   c[1]=pp[1]+mu*n[1];
   c[2]=pp[2]+mu*n[2];
   VECSUB(v,c,dn)
   moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
   rmu=sqrt(rad*rad-mu*mu);
   if(moddn < rmu)
   {
     a=1.0-moddn/rmu;
     normalize(dn);
     //
     // For Dimples the surface normal must be bent in the oppostion
     // direction.
     //
     VECSCALE(-1.0,dn,dn);
     if(type == 0)a=sin(a*1.57079);
     else if(type == 1)a=0.5;          // pits
     else
     {
       if(a < 0.5)
       {                              // crater
         a=1.0-sigmoid(a,0.0,0.5);
       }
       else a=1;
     }
     a=1.0-(1.0-a)*depth;
     VECSUM((1.0-a)*dn,(a)*n,n)
     normalize(n);
   }
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Hexagonal (honeycomb) mesh

BOOL HexagonalMesh_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 int   type  = 1;            // bump in or out
 float thickness = 0.15;     // thickness of flat ridge
 float inner = 0.2;          // thickness of flat ridge and sloping side
 float slope = 0.5;          // ridge slope
 VECTOR c1={1.0,0.0,0.0}; // ridge color
 ////////////////////////////////////
 VECTOR pp[35]={{0.25,0.0,      0.0},      //
                {0.75,0.0,      0.0},      // Array of HCP spheres
                {0.0, ROOT32/2, 0.0},      //
                {0.5, ROOT32/2, 0.0},
                {1.0, ROOT32/2, 0.0},
                {0.25,ROOT32,   0.0},
                {0.75,ROOT32,   0.0},
                {-0.25,0.0,     0.0},
                { 1.25,0.0,     0.0},
                {-0.25,ROOT32,  0.0},
                { 1.25,ROOT32,  0.0},
                {-0.25,0.0,    ROOT32},
                { 1.25,0.0,    ROOT32},
                {-0.25,ROOT32, ROOT32},
                { 1.25,ROOT32, ROOT32},
                {-0.25,1.0/2.0/ROOT3,ROOT32/2.0},
                { 1.75,1.0/2.0/ROOT3,ROOT32/2.0},
                {0.0,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},
                {0.5,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},
                {1.0,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},
                {0.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.25,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.75,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.25,0.0,      ROOT32},
                {0.75,0.0,      ROOT32},
                {0.0, ROOT32/2, ROOT32},
                {0.5, ROOT32/2, ROOT32},
                {1.0, ROOT32/2, ROOT32},
                {0.25,ROOT32,   ROOT32},
                {0.75,ROOT32,   ROOT32}};
 float dpoint[35];
 long i;
 float a,cc,r,g,b;
 VECTOR u,v;
 VECTOR dn,vt;
 float dx,dy,dz,zb,ccc,dc,zbb;
 BOOL ins;
 int id;
 cc=0.85;                 // Radius of spheres so that they overlap
 //
 //  Scale for unit cell size.
 //
 yy /= ROOT32;
 xx=FMOD(xx,1.0);
 yy=FMOD(yy,ROOT32);
 zz /= ROOT32;
 zz=FMOD(zz,ROOT32);
 zb=-10.0;
 VECCOPY(n,dn)   // temporary copy of surface normal
 r=color[0];     //
 g=color[1];     //  Default color
 b=color[2];     //

 ins=TRUE;       // Flag will be true on the sides and top of ridge

 for(i=0;i<35;i++)
 {               // Test all spheres
   dpoint[i] = -1.0;
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if(fabs(dx) > cc)continue;
   if(fabs(dy) > cc)continue;
   if(fabs(dz) > cc)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc)
   {
     dpoint[i]=a;
     dc=(cc-a);
     if(dc > zb){
       id=i;          //  Use "Z" buffer to record which is closest to point
       zb=dc;         //  "p" (xx,yy,zz)
     }
   }
 }
 zbb= -10.0;
 for(i=0;i<35;i++)
 {
   if(i == id)continue;
   if(dpoint[i] >= 0.0)
   {
     dc=(cc-dpoint[i]);
     if(dc > zbb){    // get the distance from the nearest of the other
       zbb=dc;        // spheres
     }
   }
 }
 //
 // If the difference (zbb-zb) < a threshold value then the point (xx,yy,zz)
 // lies within a small distance of the line of intersection where
 // adjacent spheres overlap.
 //
 if((ccc=fabs(zbb - zb)) < thickness)ins=FALSE;   // flat edge
 if(ins && ccc < inner)
 {                    // On the ridge slopes => the surface normal
                      // must be bent. Use an approximation by assuming
                      // the slope is that of a truncated dimple.
   v[0]=(xx-pp[id][0]);  //
   v[1]=(yy-pp[id][1]);  // Vector from (xx,yy,zz) to center of sphere
   v[2]=(zz-pp[id][2]);  //
   VECCOPY(dn,n)
   normalize(v);
   VECSCALE(v[0],U,vt)    //
   VECSUM( v[1]*V,vt,vt)  //  Vector "v" transformed to global reference
   VECSUM(-v[2]*W,vt,vt)  //
   normalize(vt);
   VECCOPY(vt,v)
   CROSS(v,n,u)
   CROSS(n,u,v)           // "v" is made perpendicular to "n"
   normalize(v);
   a=slope;
   if(type == 0)VECSUM(-a*v,(1.0-a)*n,n)
   else         VECSUM(a*v,(1.0-a)*n,n)
   normalize(n);
 }
 if(!ins)       // set the "NOT" top of ridge to a different color
 {
   r=c1[0];
   g=c1[1];
   b=c1[2];
 }
 color[0]  = r;
 color[1]  = g;
 color[2]  = b;

 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Contours  - basically a BOZO texture with noise value functionally
//              processed

BOOL Contours_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 float thickness = 0.15;     // thickness of contour
 VECTOR c1={0.0,0.0,0.0};    // "black" contours
 ////////////////////////////////////
 float noiseH;
 noiseH = fNoise(xx, yy, zz);
 //
 // pass the noise value through a sinusoidal filter.
 //
 noiseH = sintab(noiseH*10.0);
 if(noiseH > 1.0-thickness)
 {
   color[0]=noiseH*c1[0];
   color[1]=noiseH*c1[1];
   color[2]=noiseH*c1[2];
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Plasma - process the "Noise function"

BOOL Plasma_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters //////
 float brite=1.0;                                 // brightness
 float red_freq=1.0,green_freq=2.0,blue_freq=3.0; // frequencies
 float red_phas=0.0,green_phas=PI/2,blue_phas=PI; // phases
 ////////////////////////////////////
 float noiseH,r,g,b;
 //
 // Sum 3 noise functions at different scales
 //
 noiseH  = fNoise(xx, yy, zz);
 noiseH += fNoise(xx*0.5, yy*0.5, zz*0.5);
 noiseH += fNoise(xx*0.25, yy*0.25, zz*0.25);
 r = costab(noiseH*red_freq+red_phas);
 g = costab(noiseH*green_freq+green_phas);
 b = costab(noiseH*blue_freq+blue_phas);
 color[0]  = r * brite;
 color[1]  = g * brite;
 color[2]  = b * brite;
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Underwater  - another example of a processed noise (Triangle function)

BOOL Underwater_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters //////
 float coverage=0.8;        // how much of the surface has caustics
 float sharpness=0.0;       // sharply edged features
 VECTOR c1={1.0,1.0,1.0};   // caustic color
 ////////////////////////////////////
 float noiseH;
 noiseH = fNoise(xx*5.0, yy*5.0, zz*5.0);
 noiseH = triangle(noiseH);
 noiseH = (0.3+0.9*noiseH);
 if(noiseH > coverage)
 {
   noiseH=(noiseH-coverage)/(1.2-coverage);
   noiseH=noiseH*(1.0+1.0*sharpness);
   noiseH=min(noiseH,1.0);
   //
   // make the "non" caustic area  "semi transparent"
   //
   if(noiseH < 0.98)
   {
     if(noiseH < 0.5)*opacity=0.0;
     else *opacity=(noiseH-0.5)*2.084;
   }
   else *opacity=1.0;
   color[0]=(noiseH*c1[0]+(1.0-noiseH)*color[0]);
   color[1]=(noiseH*c1[1]+(1.0-noiseH)*color[1]);
   color[2]=(noiseH*c1[2]+(1.0-noiseH)*color[2]);
 }
 return TRUE;
}
