///////////////////////////////////////////////////////////////////////////
// Other worlds - simulate the appearance of a GAS giant planet with
//                bands of color

BOOL _Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 float local_noise = 0.0;   // local surface disturbances
 float filaments=0.3;       // inter band wisps
 float bands=1.0;           // how many bands
 float sharpness=1.0;       // high contrast
 VECTOR c1={1.0,0.0,0.0};   // red colored bands
 ////////////////////////////////////
 float noise1,noise2,s;
 //
 // use turbulence to add filaments or inter band mixing
 //
 noise2=fTurbulence(xx*20,yy*20,zz*10);    // filaments
 zz += noise2*filaments;                   // filaments
 //
 // differential scale for banding in vertical direction
 //
 s=fNoise(xx*0.001,yy*0.001,zz*bands);     // bands
 //
 s=s+local_noise*(fNoise(xx*50,yy*50,zz*25)-0.5);  // local noise
 s=max(s,0.0); s=min(s,1.0);
 s=0.5+(s-0.5)*sharpness;                  // sharpness
 color[0] = s*color[0]+(1-s)*c1[0];
 color[1] = s*color[1]+(1-s)*c1[1];
 color[2] = s*color[2]+(1-s)*c1[2];
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// A color fade of the Bozo texture is all that is required to
// simulate the appearance of clouds

BOOL Cloud_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 float turb=0.1;
 ////////////////////////////////////
 float noise,fraction;
 VECTOR BozoTurbulence;
 if(turb != 0.0)
 {
   //
   // distort the position to give a ragged edge to clouds
   //
   vTurbulence(BozoTurbulence, xx, yy, zz);
   xx += BozoTurbulence[0] * turb;
   yy += BozoTurbulence[1] * turb;
   zz += BozoTurbulence[2] * turb;
 }
 noise = fNoise(xx, yy, zz);
 if (noise < 0.5)return TRUE;  // basic color
 else if (noise < 0.6)
 {                             // blend to white
    fraction=(noise - 0.5)/(0.6 - 0.5);
    color[0] = color[0] + fraction*(1.0 - color[0]);
    color[1] = color[1] + fraction*(1.0 - color[1]);
    color[2] = color[2] + fraction*(1.0 - color[2]);
 }
 else
 {                             // blend to grey
    fraction=1.0 - 0.5*(noise - 0.6)/(1.0 - 0.6);
    color[0] = fraction;
    color[1] = fraction;
    color[2] = fraction;
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Leopard  - Warp an HCP dot texture to give the appearance of animal
//            skin patterns.

BOOL Leopard_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 VECTOR c1={0.8,0.1,0.0};    // pattern color
 ////////////////////////////////////                     // the location
 VECTOR pp[22]={{0.25,0.0,      0.0},                     // of HCP dots
                {0.75,0.0,      0.0},
                {0.0, ROOT32/2, 0.0},
                {0.5, ROOT32/2, 0.0},
                {1.0, ROOT32/2, 0.0},
                {0.25,ROOT32,   0.0},
                {0.75,ROOT32,   0.0},
                {0.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.25,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.75,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.25,0.0,      ROOT32},
                {0.75,0.0,      ROOT32},
                {0.0, ROOT32/2, ROOT32},
                {0.5, ROOT32/2, ROOT32},
                {1.0, ROOT32/2, ROOT32},
                {0.25,ROOT32,   ROOT32},
                {0.75,ROOT32,   ROOT32}};                 //21
 static float cc=0.25;
 float dx,dy,dz,zb;
 long i;
 float a,r,g,b;
 //
 // Warp the texture appearance by distorting the input coordinate with
 // periodic functions.
 //
 xx += 0.11*sin(FMOD(3*zz,2*3.1415926));
 xx += 0.2*sin(FMOD(0.1*yy,2*3.1415926));
 yy += 0.15*cos(FMOD(4*xx,2*3.1415926));
 yy += 0.2*sin(FMOD(0.1*zz,2*3.1415926));
 zz += 0.18*cos(FMOD(5*yy,2*3.1415926));
 zz += 0.2*sin(FMOD(0.1*xx,2*3.1415926));
 //
 // For non-unit texture cell
 //
 yy /= ROOT32;
 xx=FMOD(xx,1.0);
 yy=FMOD(yy,ROOT32);
 zz /= ROOT32;
 zz=FMOD(zz,ROOT32);
 zb=0.0;
 r=color[0];
 g=color[1];
 b=color[2];
 for(i=0;i<22;i++)         //  check each texture sphere
 {
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if(fabs(dx) > cc)continue;
   if(fabs(dy) > cc)continue;
   if(fabs(dz) > cc)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc)
   {
     if((cc-a) > zb){      // if "p" lies inside texture sphere
       zb=(cc-a);
       r=c1[0];
       g=c1[1];
       b=c1[2];
       if(a > 0.20)        // give soft edges to texture features
       {
         mix(&r,&g,&b,(a-0.20)/(cc-0.20),
          (float)color[0],
          (float)color[1],
          (float)color[2],
          r,g,b);
       }
     }
   }
 }
 color[0]  = r;
 color[1]  = g;
 color[2]  = b;
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Reptile  - rather like leopard but with a series of overlapping
//            closely spaced pimples

BOOL Reptile_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 VECTOR c1={1.0,1.0,0.0};  // color of scales
 VECTOR c2={0.0,0.0,0.0};  // scale edge color
 ////////////////////////////////////
 VECTOR pp[22]={{0.25,0.0,      0.0},     // the location of the spheres
                {0.75,0.0,      0.0},
                {0.0, ROOT32/2, 0.0},
                {0.5, ROOT32/2, 0.0},
                {1.0, ROOT32/2, 0.0},
                {0.25,ROOT32,   0.0},
                {0.75,ROOT32,   0.0},
                {0.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.25,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.75,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.25,0.0,      ROOT32},
                {0.75,0.0,      ROOT32},
                {0.0, ROOT32/2, ROOT32},
                {0.5, ROOT32/2, ROOT32},
                {1.0, ROOT32/2, ROOT32},
                {0.25,ROOT32,   ROOT32},
                {0.75,ROOT32,   ROOT32}};
 VECTOR dn,vt;
 float dx,dy,dz,zb,ccc;
 long i;
 float a,cc,r,g,b;
 VECTOR u,v;

 cc=0.5;    // sphere radius that is large enough that the dots overlap

 xx += 0.16*sin(FMOD(2*zz,2*3.1415926));  // warp the coodinates
 yy += 0.16*cos(FMOD(3*xx,2*3.1415926));  // to give ripple effect
 zz += 0.16*cos(FMOD(4*yy,2*3.1415926));  //

 yy /= ROOT32;                // HCP texture cell size scaling
 xx=FMOD(xx,1.0);
 yy=FMOD(yy,ROOT32);
 zz /= ROOT32;
 zz=FMOD(zz,ROOT32);
 zb=0.0;

 r=c1[0];   // color of scales
 g=c1[1];
 b=c1[2];
 VECCOPY(n,dn)
 for(i=0;i<22;i++)       // Test for (xx,yy,zz) lying inside sphere
 {
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc)
   {
     if((cc-a) > zb)     // use the most prominent sphere (Z buffer)
     {
       zb=(cc-a);
       r=c2[0];
       g=c2[1];
       b=c2[2];
       if(a > 0.15)
       {
         mix(&r,&g,&b,(a-0.15)/(cc-0.15),
          c1[0],c1[1],c1[2],r,g,b);
       }
       v[0]=dx; v[1]=dy; v[2]=dz;
       VECCOPY(dn,n)            // Calculate perturbing vector for
       normalize(v);            // surface normal with reference
       VECSCALE(v[0],U,vt)      // to the global frame of reference
       VECSUM(v[1]*V,vt,vt)     //
       VECSUM(-v[2]*W,vt,vt)
       normalize(vt);
       VECCOPY(-vt,v)
       ccc=fabs(DOT(v,n));
       CROSS(v,n,u)             //
       CROSS(n,u,v)             // Perturbing vector perpendicular to "n"
       normalize(v);            //
       a=1.8*a*sqrt(1.0-ccc*ccc);
       VECSCALE(-a,v,v)                    // bend for pimple
       VECSCALE((1.0-a),n,n)
       VECSUM(v,n,n)
       normalize(n);
     }
   }
 }
 color[0]  = r;
 color[1]  = g;
 color[2]  = b;
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Veins  - based on the Hexagonal Mesh

BOOL Veins_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 VECTOR c1={1.0,0.0,0.0};  //vein color
 ////////////////////////////////////
 VECTOR pp[35]={{0.25,0.0,      0.0},
                {0.75,0.0,      0.0},
                {0.0, ROOT32/2, 0.0},
                {0.5, ROOT32/2, 0.0},
                {1.0, ROOT32/2, 0.0},
                {0.25,ROOT32,   0.0},
                {0.75,ROOT32,   0.0},

                {-0.25,0.0,     0.0},
                { 1.25,0.0,     0.0},
                {-0.25,ROOT32,  0.0},
                { 1.25,ROOT32,  0.0},
                {-0.25,0.0,    ROOT32},
                { 1.25,0.0,    ROOT32},
                {-0.25,ROOT32, ROOT32},
                { 1.25,ROOT32, ROOT32},
                {-0.25,1.0/2.0/ROOT3,ROOT32/2.0},
                { 1.75,1.0/2.0/ROOT3,ROOT32/2.0},
                {0.0,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},
                {0.5,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},
                {1.0,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},

                {0.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.25,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.75,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.25,0.0,      ROOT32},
                {0.75,0.0,      ROOT32},
                {0.0, ROOT32/2, ROOT32},
                {0.5, ROOT32/2, ROOT32},
                {1.0, ROOT32/2, ROOT32},
                {0.25,ROOT32,   ROOT32},
                {0.75,ROOT32,   ROOT32}};
 VECTOR u,v;
 static float cc=0.4;
 float a;
 float dx,dy,dz,zb,ccc,dc,zbb;
 BOOL ins;
 int i,id;

 vNoise(v,xx,yy,zz);   // Distort input coordinate with a Noise
 xx += 0.7*v[0];
 yy += 0.7*v[1];
 zz += 0.7*v[2];

 yy /= ROOT32;         // adjust for "non unit" unit cell size
 xx=FMOD(xx,1.0);
 yy=FMOD(yy,ROOT32);
 zz /= ROOT32;
 zz=FMOD(zz,ROOT32);
 zb=-10.0;

 ins=TRUE;
 for(i=0;i<35;i++)     // test each sphere
 {
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc)
   {
     dc=(cc-a);
     if(dc > zb){
       id=i;              // identify closest
       zb=dc;
     }
   }
 }
 //
 //  get distance from next closest
 //
 zbb= -10.0;
 for(i=0;i<35;i++)
 {
   if(i == id)continue;
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc)
   {
     dc=(cc-a);
     if(dc > zbb){
       zbb=dc;
     }
   }
 }
 //
 // See hexagonal mesh code for explanation
 // ins = FALSE if "p" lies inside a vein
 //
 if((ccc=fabs(zbb - zb)) < 0.05)ins=FALSE;
 if(!ins)
 {
   v[0]=(xx-pp[id][0]);     // bend the normal in the veins to give
   v[1]=(yy-pp[id][1]);     // a gouged appearance
   v[2]=(zz-pp[id][2]);
   normalize(v);
   CROSS(v,n,u)
   CROSS(n,u,v)
   normalize(v);
   a=0.4;
   VECSCALE(a,v,v)
   VECSCALE((1.0-a),n,n)
   VECSUM(v,n,n)
   normalize(n);
   color[0]=c1[0];         // vein color
   color[1]=c1[1];
   color[2]=c1[2];
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Random dots and Crystals

BOOL RandomDots_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 int rtype=1;        // type? dots or rings
 float bradius=0.2;  // Radius of dots (if > 0.5 then crystals
                     // because dots interlock together.)
 float rradius=0.25; // Fraction of radius for inner ring radius
 float lradius=0.0;  // Dispsersion in size of dots
 ////////////////////////////////////
 float noiseH,r,g,b;
 VECTOR  vclr[27];
 float dradius[27];
 VECTOR u,v,pp[10];
 float a,dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius;
 long ix,iy,iz;
 long i,j,jid,id,pattern,patterni;
 long N=1;
 //
 // make (xx,yy,zz) relative to the unit cell: (cx,cy,cz) is the result
 //
 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(float)ix;
 cy=yy-(float)iy;
 cz=zz-(float)iz;
 zb=-10.0;
 //
 // Test for containment inside the sphere in the unit cell based on
 // (ix,iy,iz) and its 26 neighbors.
 //
 id = -1;
 i=0;           // only one sphere per unit cell
 for(j=0;j<27;j++)
 {
   //
   // Use a hashing function to address a table of random vectors based
   // on which cell we are considering
   //
   pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
   pattern=IMOD(N*pattern,1900);
   i=0;
   jid=j*N+i;
   patterni=pattern+i;
   VECSUM((float)ip[j],rand_v[patterni],pp[i])
   //
   // choose a random color for dot "jid"
   //
   VECCOPY(rand_v[patterni],vclr[jid])
   //
   // vairy the radius of the dot
   //
   radius=bradius+lradius*(rand_v[patterni][1]-0.5);
   dradius[jid]=radius;
   dx=(cx-pp[i][0]);
   dy=(cy-pp[i][1]);
   dz=(cz-pp[i][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius)
   {                             // If the point lies inside sphere "jid"
                                 // With just one sphere per cell "jid"
     dc=(radius-a);              // really identifies which of the 27
     if(dc > zb){                // cells the sphere lies in
       id=jid;
       zb=dc;
     }
   }
 }
 //
 //  if "id" is non zero then "p" lies inside the sphere identified by "id"
 //
 if(id >= 0
   && (rtype == 0 || ( zb < dradius[id]-rradius*dradius[id] ) ) // rings
   ){
   r=vclr[id][0];     //
   g=vclr[id][1];     //  Get the random color of the dot we can see
   b=vclr[id][2];     //  at "p"
   if(zb < 0.1)
   {
     mix(&r,&g,&b,    // Give "soft" edge to the dot
         zb/(0.1),
         r,g,b,
         color[0],
         color[1],
         color[2]
         );
   }
 }
 else{
   r=color[0];
   g=color[1];
   b=color[2];
 }
 color[0]  = r;
 color[1]  = g;
 color[2]  = b;
 return TRUE;
}
