///////////////////////////////////////////////////////////////////////////
// Blisters  - Based on the random dots and reptile bumps
//             (see those textures for more details)

BOOL Blisters_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 int type = 1;       // shape of bump
 float depth=1.0;    // strength of the bump
 float size=0.2;     // blister radius
 VECTOR c1={0.1,0.1,0.0};
 ////////////////////////////////////
 float a;
 VECTOR u,v;
 float dradius[27];
 VECTOR pp[27];
 float dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius;
 long ix,iy,iz;
 long i,j,jid,id,pattern,patterni;
 static long N=1;
 static float lradius=0.01;
 float rad;
 VECTOR ppc;

 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(float)ix;
 cy=yy-(float)iy;
 cz=zz-(float)iz;
 zb=-10.0;
 id = -1;
 VECCOPY(P,v)
 VECSUM(v,cx*U,v)
 VECSUM(v,cy*V,v)
 VECSUM(v,cz*W,v)         // coord of point on surface

 for(j=0;j<27;j++)        // test random sphere in cell "i" and its 26
 {                        // adjacent cells
   pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
   pattern=IMOD(N*pattern,1900);
   i=0;
   jid=j*N+i;
   patterni=pattern+i;
   //
   //  Test for intersection in global coordinate frame.
   //
   VECSUM((float)ip[j],rand_v[patterni],ppc)
   VECSUM(P,ppc[0]*U,pp[jid])
   VECSUM(pp[jid],ppc[1]*V,pp[jid])
   VECSUM(pp[jid],ppc[2]*W,pp[jid])
   radius=(size+lradius*(rand_v[patterni][1]-0.5));
   dradius[jid]=radius;
   dx=(v[0]-pp[jid][0]);
   dy=(v[1]-pp[jid][1]);
   dz=(v[2]-pp[jid][2]);
   if(fabs(dx) > radius)continue;  //    Not this one !
   if(fabs(dy) > radius)continue;  //         " "
   if(fabs(dz) > radius)continue;  //         " "
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius)
   {
     dc=(radius-a);
     if(dc > zb)    // Record identity of which cell sphere point
     {              // "p" lies inside
       id=jid;
       zb=dc;
     }
   }
 }
 if(id >= 0)                  // inside cell "id"
 {
     //
     // Use the procedure of the pimple/dimple textures to calculate
     // the displacement applied to the surface normal based on
     // how far "p" is from the center of the sphere which in turn is
     // closest to "p".
     //
     VECCOPY(pp[id],ppc)
     rad=dradius[id];
     VECSUB(v,ppc,u)             //(v-pp) VECTOR from centre of pattern
     if((a=sqrt((u[0])*(u[0])+
                (u[1])*(u[1])+
                (u[2])*(u[2]))) < rad)
     {
       float mu,moddn,rmu;
       VECTOR c,dn;
       mu=DOT(u,n);
       c[0]=ppc[0]+mu*n[0];
       c[1]=ppc[1]+mu*n[1];
       c[2]=ppc[2]+mu*n[2];
       VECSUB(v,c,dn)
       moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
       rmu=sqrt(rad*rad-mu*mu);
       if(moddn < rmu)
       {
         a=1.0-moddn/rmu;
         normalize(dn);
         //
         // modify the magnitude of the "normal's"  displacement vector
         //
         if(type == 0)a=sin(a*1.57079);
         else{
           if(type == 1)a=pow(a,0.25);        // flatter bumps
           else
           {
             if(a > 0.5)                      // flat colored centers
             {
               a=1.0;
               if(a > 0.6)VECCOPY(c1,color)
             }
             else a=pow(a,0.25);
           }
         }
         a=1.0-(1.0-a)*depth;                // reduced bump size
         VECSUM((1.0-a)*dn,(a)*n,n)
         normalize(n);
         if(type != 2)VECCOPY(c1,color)
       }
   }
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Peas  - is very similar to the Blisters texture

BOOL Peas_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 int type=1;          // choose model
 int flatter=1;       // flatter peas
 ////////////////////////////////////
 static long N1=3;
 static float bradius1=0.55,lradius1=0.2;
 static long N2=1;
 static float bradius2=1.2,lradius2=1.0;
 long i,N;
 float a;
 VECTOR u,v;
 float dradius[81];  // up to 81 spheres are considered (3 per cubic latice)
 VECTOR pp[81];
 float dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius,bradius,lradius;
 long ix,iy,iz;
 long j,jid,id,pattern,patterni;
 float rad;
 VECTOR ppc;
 if(type == 1)   // full coverage
 {
   bradius=bradius2;
   lradius=lradius2;
   N=N2;
 }
 else            // Partial coverage, 3 random spheres are in each unit cell
 {
   bradius=bradius1;
   lradius=lradius1;
   N=N1;
 }
 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(float)ix;
 cy=yy-(float)iy;
 cz=zz-(float)iz;
 zb=-10.0;
 id = -1;
 VECCOPY(P,v)
 VECSUM(v,cx*U,v)
 VECSUM(v,cy*V,v)
 VECSUM(v,cz*W,v)  // coordinate of point on surface w.r.t. global coordinates
 //
 // Examine the spheres in each texture cell adjacent to cell (ix,iy,iz)
 //
 for(j=0;j<27;j++)
 {
 pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
 pattern=IMOD(N*pattern,1900);
 for(i=0;i<N;i++)
 {
   //
   // Get the coords and radius of sphere "i" in cell "j". In cell "j"
   // the sphere is positioned randomly by choosing a vector from a table
   // of random vectors.  A hashing table returns a consistant random
   // vector given a known (ix,iy,iz).
   //
   jid=j*N+i;
   patterni=pattern+i;
   VECSUM((float)ip[j],rand_v[patterni],ppc)
   VECSUM(P,ppc[0]*U,pp[jid])
   VECSUM(pp[jid],ppc[1]*V,pp[jid])
   VECSUM(pp[jid],ppc[2]*W,pp[jid])
   radius=(bradius+lradius*(rand_v[patterni][1]-0.5));
   dradius[jid]=radius;
   dx=(v[0]-pp[jid][0]);
   dy=(v[1]-pp[jid][1]);
   dz=(v[2]-pp[jid][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius)
   {
     dc=(radius-a);
     if(dc > zb){
       id=jid;   // record the "id" of closest sphere to "p"  i.e. "v"
       zb=dc;
     }
   }
 }
 }
 if(id >= 0)
 {
     //
     // Use the procedure of the pimple/dimple textures to calculate
     // the displacement applied to the surface normal based on
     // how far "p" is from the center of the sphere which in turn is
     // closest to "p".
     //
     VECCOPY(pp[id],ppc)
     rad=dradius[id];
     VECSUB(v,ppc,u)          //(v-pp) VECTOR from centre of pattern
     if((a=sqrt((u[0])*(u[0])+
                (u[1])*(u[1])+
                (u[2])*(u[2]))) < rad)
     {
       float mu,moddn,rmu;
       VECTOR c,dn;
       mu=DOT(u,n);
       c[0]=ppc[0]+mu*n[0];
       c[1]=ppc[1]+mu*n[1];
       c[2]=ppc[2]+mu*n[2];
       VECSUB(v,c,dn)
       moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
       rmu=sqrt(rad*rad-mu*mu);
       if(moddn < rmu)
       {
         a=1.0-moddn/rmu;
         normalize(dn);
         if(flatter == 1)a=sin(a*1.57079);
         VECSUM((1.0-a)*dn,(a)*n,n)
         normalize(n);
       }
     }
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Honeycomb - regular hexagonal mesh with dents in the center

BOOL Honeycomb_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 int type=2;                    // How color is mixed
 int ifade=0;                   //    "      "
 VECTOR c1={255.0,255.0, 0.0};  // Edge color
 float thickness=0.05;          // Width of hexagonal edge
 float coverage=0.05;           // Fractions over which
 float solidsize=0.0;           // the color is mixed.
 ////////////////////////////////////
 VECTOR pp[35]={{0.25,0.0,      0.0},             // the HCP spheres
                {0.75,0.0,      0.0},             // forming the dimples
                {0.0, ROOT32/2, 0.0},
                {0.5, ROOT32/2, 0.0},
                {1.0, ROOT32/2, 0.0},
                {0.25,ROOT32,   0.0},
                {0.75,ROOT32,   0.0},

                {-0.25,0.0,     0.0},
                { 1.25,0.0,     0.0},
                {-0.25,ROOT32,  0.0},
                { 1.25,ROOT32,  0.0},
                {-0.25,0.0,    ROOT32},
                { 1.25,0.0,    ROOT32},
                {-0.25,ROOT32, ROOT32},
                { 1.25,ROOT32, ROOT32},
                {-0.25,1.0/2.0/ROOT3,ROOT32/2.0},
                { 1.75,1.0/2.0/ROOT3,ROOT32/2.0},
                {0.0,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},
                {0.5,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},
                {1.0,1.0/2.0/ROOT3+ROOT32,ROOT32/2.0},

                {0.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3-ROOT32/2,ROOT32/2.0},
                {0.25,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.75,1.0/2.0/ROOT3,         ROOT32/2.0},
                {0.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.5, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {1.0, 1.0/2.0/ROOT3+ROOT32/2,ROOT32/2.0},
                {0.25,0.0,      ROOT32},
                {0.75,0.0,      ROOT32},
                {0.0, ROOT32/2, ROOT32},
                {0.5, ROOT32/2, ROOT32},
                {1.0, ROOT32/2, ROOT32},
                {0.25,ROOT32,   ROOT32},
                {0.75,ROOT32,   ROOT32}};
 VECTOR u,v;
 float cc=0.85;  // radius of sphere to ensure overlapping spheres
 float a,aa,as;
 float dx,dy,dz,zb,ccc,dc,zbb;
 BOOL ins;
 int i,id;
 //
 // Adjust the aspect ration of the texture cell to give the illusion
 // of being a unit cube.
 //
 yy /= ROOT32;
 xx=FMOD(xx,1.0);
 yy=FMOD(yy,ROOT32);
 zz /= ROOT32;
 zz=FMOD(zz,ROOT32);
 zb=-10.0;
 ins=TRUE;

 for(i=0;i<35;i++)   // test each sphere to find the closest to "p"
 {
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc)
   {
     dc=(cc-a);
     if(dc > zb){
       id=i;      // record the "id" of the closest
       zb=dc;
       as=a;
     }
   }
 }
 //
 // Get the difference between the closest and the next closest
 // this gives is a way of establishing a border round the edges
 // of the hexagons which is parallel to the edges of the hexagons
 //
 zbb= -10.0;
 for(i=0;i<35;i++)
 {
   if(i == id)continue;
   dx=(xx-pp[i][0]);
   dy=(yy-pp[i][1]);
   dz=(zz-pp[i][2]);
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < cc){
     dc=(cc-a);
     if(dc > zbb){
       zbb=dc;
     }
   }
 }
 //
 // See hexagonal mesh code for explanation
 // ins = FALSE if "p" lies inside the hexagonal edge
 //
 if((ccc=fabs(zbb - zb)) < thickness)ins=FALSE;
 if(ins)
 {   //
     //  The inside of the hexagons which is a simple dimple
     //  with its associated "bend" in the surface normal.
     //
     v[0]=(xx-pp[id][0]);
     v[1]=(yy-pp[id][1]);
     v[2]=(zz-pp[id][2]);   // a vector pointing from center of sphere to
     normalize(v);          // point of interest on the surface.
     CROSS(v,n,u)
     CROSS(n,u,v)           // v is now a vector perpendicular to "n"
     normalize(v);
     aa=0.4;
     VECSCALE(aa,v,v)
     VECSCALE((1.0-aa),n,n)
     VECSUM(v,n,n)
     normalize(n);
 }
 //
 // Assorted color fades from inside to outside or over the edge,
 // depending in the value of type and the other parameters/
 //
 if(type > 0)
 {
   if(type == 2){        // fade away from edge
     if(ccc < coverage)
     {                   // distance from hexagonal edge
       if(ifade && ccc > solidsize)
       {
         mix(&color[0],&color[1],&color[2],
           (ccc-solidsize)/(coverage-solidsize),
           color[0],
           color[1],
           color[2],
           c1[0],c1[1],c1[2]);
       }
       else{
         color[0]=c1[0];
         color[1]=c1[1];
         color[2]=c1[2];
       }
     }
   }
   else
   { // fade from center
     if(as < coverage)
     {
       if(ifade && as > solidsize)
       {
         mix(&color[0],&color[1],&color[2],
           (as-solidsize)/(coverage-solidsize),
           color[0],
           color[1],
           color[2],
           c1[0],c1[1],c1[2]);
       }
       else
       {
         color[0]=c1[0];
         color[1]=c1[1];
         color[2]=c1[2];
       }
     }
   }
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Paving stones  -  A mixture of the Crystalized random dots with
//                   bevelled edge cracks formed in the same way as the
//                   bevelled edges of the Vein texture. See the code of
//                   those textures for algorithmic information.

BOOL Paving_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 float flat_edge=0.025;      // width of floor of crack
 float bevel_edge=0.06;      // width of side of crack
 float blend_edge=0.06;      // where color blend starts
 VECTOR c1={0.5,0.3,0.1};    // color of crack
 ////////////////////////////////////
 float r,g,b;
 float a;
 VECTOR u,v;
 VECTOR vclr[27];
 float dpoint[27];
 VECTOR pp[10];
 float dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,noiseH;
 long ix,iy,iz;
 BOOL ins;
 long i,j,jid,id,pattern,patterni;

 float radius=1.0;       // to ensure "crystalization" of spheres
 float vc=0.1;           // discoloration

 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(float)ix;
 cy=yy-(float)iy;
 cz=zz-(float)iz;
 zb=-10.0;
 ins=TRUE;

 for(i=0,j=0;j<27;j++)  // try one sphere per neghboring texture cell
 {
   //
   //  address the random position vector for sphere "i"
   //
   pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
   pattern=IMOD(pattern,1900);
   jid=j+i;
   patterni=pattern+i;
   VECSUM((float)ip[j],rand_v[patterni],pp[i])
   VECCOPY(rand_v[patterni],vclr[jid])
   dpoint[jid] = -1.0;
   dx=(cx-pp[i][0]);
   dy=(cy-pp[i][1]);
   dz=(cz-pp[i][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius)  // inside sphere
   {
     dpoint[jid]=a;
     dc=(radius-a);
     if(dc > zb)
     {
       id=jid;         // record "id" of sphere to be used
       zb=dc;
     }
   }
 }
 //
 // find next closest and record its distance
 //
 zbb= -10.0;
 for(i=0;i<27;i++)
 {
   if(i == id)continue;
   if(dpoint[i] >= 0.0){
     dc=(radius-dpoint[i]);
     if(dc > zbb){
       zbb=dc;
     }
   }
 }
 //
 // is it in the crack ?
 //
 if((ccc=fabs(zbb - zb)) < flat_edge)ins=FALSE;
 if(ins)
 {
   //
   //  Non on the bottom of the crack - either on the sloping side
   //  or in the main body of stone "id"
   //
   if(0)
   {
     r=vclr[id][0];   // totally random colored paving  (crystals)
     g=vclr[id][1];
     b=vclr[id][2];
   }
   else
   {
     //
     //  Randomly discolor each flag stone to give variation in pattern
     //
     r=color[0]/255.0 + vc*(vclr[id][0]-0.5);
     g=color[1]/255.0 + vc*(vclr[id][1]-0.5);
     b=color[2]/255.0 + vc*(vclr[id][2]-0.5);
     r=min(r,1.0); g=min(g,1.0); b=min(b,1.0);
     //
     //  Add a small fraction of "Noise" to color  to give a weathered
     //  and dirty appearance of stones.
     //
     noiseH=fNoise(xx*2.0,yy*2.0,zz*2.0);  // 2 X frequency variation
     r -= (noiseH*0.3);
     g -= (noiseH*0.3);
     b -= (noiseH*0.3);
     r=max(r,0.0); g=max(g,0.0); b=max(b,0.0);
   }
   if(ccc < bevel_edge)
   {
     v[0]=(cx-pp[id][0]);       // On the side of the crack  - bend the
     v[1]=(cy-pp[id][1]);       // normal in the usual manner.
     v[2]=(cz-pp[id][2]);       //
     normalize(v);
     CROSS(v,n,u)
     CROSS(n,u,v)
     normalize(v);
     a=0.4;                     // slope
     VECSCALE(a,v,v)
     VECSCALE((1.0-a),n,n)      // mix
     VECSUM(v,n,n)
     normalize(n);
   }
   if(ccc < blend_edge)         // blend edge colors
      mix(&r,&g,&b,(ccc-flat_edge)/(blend_edge-flat_edge),
         r,g,b,
         (float)c1[0],
         (float)c1[1],
         (float)c1[2]
         );
 }
 else
 {
   r=c1[0];   // on the floor of hte crack
   g=c1[1];
   b=c1[2];
 }
 color[0]=r;   // return the new color
 color[1]=g;
 color[2]=b;

 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
// Pieces   - This is exactly the same as the Paving texture with the
//            exception that the cracks are very wide and the bevelled
//            edges of the cracks are much more pronounced.

BOOL Pieces_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 float flat_edge=0.085;      // width of floor of crack
 float bevel_edge=0.15;      // width of side of crack
 float blend_edge=0.15;      // where color blend starts
 VECTOR c1={0.5,0.3,0.1};    // color of crack
 ////////////////////////////////////
 float r,g,b;
 float a;
 VECTOR u,v,ppc;
 VECTOR vclr[27];
 float dpoint[27];
 VECTOR pp[27];
 float dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,noiseH;
 long ix,iy,iz;
 BOOL ins;
 long j,id,pattern,patterni;
 static float radius,bradius=1.0; // to ensure "crystallation"
 static float vc=0.2;             // random color variation

 id=-1;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(float)ix;
 cy=yy-(float)iy;
 cz=zz-(float)iz;
 zb=-10.0;
 ins=TRUE;

 VECCOPY(P,v)
 VECSUM(v,cx*U,v)
 VECSUM(v,cy*V,v)
 VECSUM(v,cz*W,v)
 radius=bradius;
 //
 // Which Sphere
 //
 for(j=0;j<27;j++)
 {
   pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
   pattern=IMOD(pattern,1900);
   patterni=pattern;

   VECSUM((float)ip[j],rand_v[patterni],ppc)

   VECSUM(P,ppc[0]*U,pp[j])
   VECSUM(pp[j],ppc[1]*V,pp[j])
   VECSUM(pp[j],ppc[2]*W,pp[j])

   VECCOPY(rand_v[patterni],vclr[j])
   dpoint[j] = -1.0;
   dx=(v[0]-pp[j][0]);
   dy=(v[1]-pp[j][1]);
   dz=(v[2]-pp[j][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius)
   {
     dpoint[j]=a;
     dc=(bradius-a);
     if(dc > zb)
     {
       id=j;      // ID of sphere
       zb=dc;
     }
   }
 }
 //
 //  To calculate crack width.
 //
 zbb= -10.0;
 for(j=0;j<27;j++)
 {
   if(j == id)continue;
   if(dpoint[j] >= 0.0){
     dc=(bradius - dpoint[j]);
     if(dc > zbb){
       zbb=dc;
     }
   }
 }
 //
 // Inside the crack ?   (near the edges of the crystals)
 //
 if((ccc=fabs(zbb - zb)) < flat_edge)ins=FALSE;
 if(id >= 0 && ins)
 {
   //
   // Discolor cracked pieces
   //
   r=color[0] + vc*(vclr[id][0]-0.5);
   g=color[1] + vc*(vclr[id][1]-0.5);
   b=color[2] + vc*(vclr[id][2]-0.5);
   r=min(r,1.0); g=min(g,1.0); b=min(b,1.0);
   noiseH=fNoise(xx*2.0,yy*2.0,zz*2.0);
   r -= (noiseH*0.13);
   g -= (noiseH*0.13);
   b -= (noiseH*0.13);
   r=max(r,0.0); g=max(g,0.0); b=max(b,0.0);
   if(ccc < bevel_edge)
   {
     //
     // Bend normal on the bevelled side of the cracked pieces
     //
     VECTOR ppc;
     VECCOPY(pp[id],ppc)
     VECSUB(v,ppc,u)
     if((a=sqrt((u[0])*(u[0])+
                (u[1])*(u[1])+
                (u[2])*(u[2]))) < radius)
     {
       //
       // Calculate the bend for the normal on the sides of the
       // cracked pieces.
       //
       float mu,moddn,rmu;
       VECTOR c,dn;
       mu=DOT(u,n);
       c[0]=ppc[0]+mu*n[0];
       c[1]=ppc[1]+mu*n[1];
       c[2]=ppc[2]+mu*n[2];
       VECSUB(v,c,dn)
       moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
       rmu=sqrt(radius*radius-mu*mu);
       if(moddn < rmu)
       {
         normalize(dn);
         dn[0]=0.25*(float)((int)(dn[0]*4.0));  //  Flatten the sides of the
         dn[1]=0.25*(float)((int)(dn[1]*4.0));  //  dimples by quantizing
         dn[2]=0.25*(float)((int)(dn[2]*4.0));  //  the vector components.
         normalize(dn);
         a=0.5;
         VECSUM((1.0-a)*dn,(a)*n,n)
         normalize(n);
       }
     }
   }
   if(ccc < blend_edge)  // blend the edge colors
     mix(&r,&g,&b,(ccc-flat_edge)/(blend_edge-flat_edge),
         r,g,b,
         (float)c1[0],
         (float)c1[1],
         (float)c1[2]
         );
 }
 else
 {
   r=c1[0];  // inside the "wide" cracks
   g=c1[1];
   b=c1[2];
 }
 color[0]=r;
 color[1]=g;
 color[2]=b;

 return TRUE;
}
