///////////////////////////////////////////////////////////////////////////
//  Rough  - a straightforward use of the "Perlin" type noise

BOOL Rough_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ////////////
 float height=1.0;   // height of the bump
 /////////////////////////////////////////
   VECTOR dr;
   vNoise(dr,xx,yy,zz);    // get the displacement vector
   VECSUM(height*dr,n,n)
   normalize(n);
   return TRUE;
}


///////////////////////////////////////////////////////////////////////////
//  Crusted  - a texture based on the ideas of the "Pea" surface. The
//             main extension is a pre-transformation of sampling point
//             (xx,yy,zz) with a "Perlin" noise.  See also "Random Dots"

BOOL Crusted_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
  int type = 3;     // type of crust
 ////////////////////////////////////
 long N=1;
 float bradius=1.2,lradius=1.0;
 float a;
 VECTOR u,v;
 float  dpoint[27];
 float dradius[27];
 VECTOR pp[27];
 float dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius;
 long ix,iy,iz;
 long i,j,jid,id,pattern,patterni;
 float rad;
 VECTOR ppc;

 //
 //  By scaling the texture and perturbing the input (sampling) point
 //  an number of "crusted" textures can be achieved.
 //

 if(type == 0 || type == 3){
   xx *= 20.0;
   yy *= 20.0;
   zz *= 20.0;
 }
 else if(type == 1){
   xx *= 4.0;
   yy *= 4.0;
   zz *= 4.0;
 }
 else{
   xx *= 4.0;
   yy *= 4.0;
   zz *= 4.0;
 }

 if(type < 3){
   vNoise(v,xx,yy,zz);   // crusted
   xx += 0.7*v[0];
   yy += 0.7*v[1];
   zz += 0.7*v[2];
 }
 //
 // Now move on the the "Pea" texture - see that code for more details
 //
 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(float)ix;
 cy=yy-(float)iy;
 cz=zz-(float)iz;
 zb=-10.0;
 id = -1;
 VECCOPY(P,v)
 VECSUM(v,cx*U,v)
 VECSUM(v,cy*V,v)
 VECSUM(v,cz*W,v)  // coord of point on surface w.r.t. unit cell in global frame

 for(j=0;j<27;j++) // the 27 nearest texture cells
 {
   pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
   pattern=IMOD(N*pattern,1900);
   i=0;
   jid=j*N+i;
   patterni=pattern+i;
   VECSUM((float)ip[j],rand_v[patterni],ppc)
   VECSUM(P,ppc[0]*U,pp[jid])
   VECSUM(pp[jid],ppc[1]*V,pp[jid])
   VECSUM(pp[jid],ppc[2]*W,pp[jid])

   radius=(bradius+lradius*(rand_v[patterni][1]-0.5));
   dradius[jid]=radius;
   dpoint[jid] = -1.0;
   dx=(v[0]-pp[jid][0]);
   dy=(v[1]-pp[jid][1]);
   dz=(v[2]-pp[jid][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius){
     dpoint[jid]=a;
     dc=(radius-a);
     if(dc > zb){
       id=jid;
       zb=dc;
     }
   }
 }
 if(id >= 0)
 {
     //
     //  Point "p" is inside sphere associated with cell "id".
     //  Bend the surface normal.
     //
     VECCOPY(pp[id],ppc)
     rad=dradius[id];
     VECSUB(v,ppc,u)
     if((a=sqrt((u[0])*(u[0])+
                (u[1])*(u[1])+
                (u[2])*(u[2]))) < rad)
     {
       float mu,moddn,rmu;
       VECTOR c,dn;
       mu=DOT(u,n);
       c[0]=ppc[0]+mu*n[0];
       c[1]=ppc[1]+mu*n[1];
       c[2]=ppc[2]+mu*n[2];
       VECSUB(v,c,dn)
       moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
       rmu=sqrt(rad*rad-mu*mu);
       if(moddn < rmu)
       {
         a=1.0-moddn/rmu;
         normalize(dn);
         //
         // The crusts are made to appear more "gnarled" by applying
         // a power law to the fraction of how near the center of sphere
         // "id" that "p" lies.
         //
         if(type == 2)a=a*a*a*a;
         else a=a*a;
         VECSUM((1.0-a)*dn,(a)*n,n)
         //
         //  Add a little "Noise" at a high frequency to superimpose
         //  some surface detail.
         //
         vNoise(v,xx*5,yy*5,zz*5);
         VECSUM(0.8*v,n,n);
         normalize(n);
       }
     }
 }
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
//  Crumpled    Add several "Pea" textures at decreasing scales

BOOL Crumples_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 ////////////////////////////////////
 int N=1;
 float bradius=1.0,lradius=0.0;
 float a;
 VECTOR u,v;
 float  dpoint[27];
 float dradius[27];
 VECTOR pp[27];
 float dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius;
 long i,ix,iy,iz;
 long k,j,jid,id,pattern,patterni;
 float rad;
 VECTOR ppc;

 for(k=0;k<3;k++)   // Repeat the texture three times at decreasing scales
 {
 if(k == 1)
 {
   //
   // After the first pass the smaller scales are distorted by
   // pre-modulating the sampling point.
   //
   vNoise(v,xx,yy,zz);
   xx += 0.5*v[0];
   yy += 0.5*v[1];
   zz += 0.5*v[2];
 }
 //
 //  Now the "pea" texture
 //
 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(float)ix;
 cy=yy-(float)iy;
 cz=zz-(float)iz;
 zb=-10.0;
 id = -1;
 VECCOPY(P,v)
 VECSUM(v,cx*U,v)
 VECSUM(v,cy*V,v)
 VECSUM(v,cz*W,v)
 //
 for(j=0;j<27;j++)   // each of the 27 neighboring texture cell
 {
   pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
   pattern=IMOD(N*pattern,1900);
   i=0;
   jid=j*N+i;
   patterni=pattern+i;
   VECSUM((float)ip[j],rand_v[patterni],ppc)
   VECSUM(P,ppc[0]*U,pp[jid])
   VECSUM(pp[jid],ppc[1]*V,pp[jid])
   VECSUM(pp[jid],ppc[2]*W,pp[jid])
   radius=(bradius+lradius*(rand_v[patterni][1]-0.5));
   dradius[jid]=radius;
   dpoint[jid] = -1.0;
   dx=(v[0]-pp[jid][0]);
   dy=(v[1]-pp[jid][1]);
   dz=(v[2]-pp[jid][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius)
   {
     dpoint[jid]=a;
     dc=(radius-a);
     if(dc > zb){
       id=jid;
       zb=dc;
     }
   }
 }
 if(id >= 0)
 {
     VECCOPY(pp[id],ppc)
     rad=dradius[id];
     VECSUB(v,ppc,u)          //(v-pp) VECTOR from centre of pattern
     if((a=sqrt((u[0])*(u[0])+
                (u[1])*(u[1])+
                (u[2])*(u[2]))) < rad)
     {
       float mu,moddn,rmu;
       VECTOR c,dn;
       mu=DOT(u,n);
       c[0]=ppc[0]+mu*n[0];
       c[1]=ppc[1]+mu*n[1];
       c[2]=ppc[2]+mu*n[2];
       VECSUB(v,c,dn)
       moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
       rmu=sqrt(rad*rad-mu*mu);
       if(moddn < rmu)
       {
         a=1.0-moddn/rmu;
         normalize(dn);
         VECSCALE(-1.0,dn,dn);                // For depressions reverse "dn"
         if     (k == 0){a=sin(a*1.57079);}   // Different attenuations
         else if(k == 1)a=0.8;                // depending on which of the
         else if(k == 2)a=0.95;               // three superimposed inverted
         VECSUM((1.0-a)*dn,(a)*n,n)           // pea textures is current
         normalize(n);
       }
     }
 }
 //
 // Increase the texture frequency (decrease the scale) after each phase
 // is superimposed.
 //
 xx *= 2.0; yy *= 2.0; zz *=2.0;

 } // end of "k" loop
 return TRUE;
}


///////////////////////////////////////////////////////////////////////////
//  Faceted -  Rather like the Crumpled texture with the important difference
//             that no attenuation is applied to each of the superimposed
//             textures and the "dimples" are quantized to have flat rather
//             than curved sides.

BOOL Faceted_Texture(
  VECTOR p,            // coordinate of surface point
  VECTOR n,            // surface normal at  "p"  (unit length)
  VECTOR P,            // position vector to origin of texture parallelepiped
  VECTOR U,            // normalized basis vectors of texture parallelepiped
  VECTOR V,
  VECTOR W,
  float  xx,           // relation between "p" and texture parallelepiped
  float  yy,           // i.e.   p = P + xx*U + yy*V + zz*W
  float  zz,
  MATRIX3 T,           // transform vector from (x,y,z) to (U,V,W) coordinate frame
  float  *opacity,     // modify the opacity of the surface if required
  float  *reflectivity,// modify the reflectivity of the surface if required
  VECTOR color         // return the surface RGB color triple (0.0 - 1.0)
)
{
 ////////  Texture Parameters ///////
 ////////////////////////////////////
 static long N=1;
 static float bradius=1.0,lradius=0.0;
 float a;
 VECTOR u,v;
 float  dpoint[27];
 float dradius[27];
 VECTOR pp[27];
 float dx,dy,dz,cx,cy,cz,zb,ccc,dc,zbb,radius;
 long i,ix,iy,iz;
 long k,j,jid,id,pattern,patterni;
 float rad;
 VECTOR ppc;

 for(k=0;k<2;k++)   // Only two superimposed "invertd pea" textures.
 {
 id=0;
 ix=FLOOR(xx);
 iy=FLOOR(yy);
 iz=FLOOR(zz);
 cx=xx-(float)ix;
 cy=yy-(float)iy;
 cz=zz-(float)iz;
 zb=-10.0;
 id = -1;
 VECCOPY(P,v)
 VECSUM(v,cx*U,v)
 VECSUM(v,cy*V,v)
 VECSUM(v,cz*W,v)
 for(j=0;j<27;j++)  // each of the 27 neighboring texture cell
 {
   pattern=PATTERN(ix+ip[j][0],iy+ip[j][1],iz+ip[j][2]);
   pattern=IMOD(N*pattern,1900);
   i=0;
   jid=j*N+i;
   patterni=pattern+i;
   VECSUM((float)ip[j],rand_v[patterni],ppc)
   VECSUM(P,ppc[0]*U,pp[jid])
   VECSUM(pp[jid],ppc[1]*V,pp[jid])
   VECSUM(pp[jid],ppc[2]*W,pp[jid])
   radius=(bradius+lradius*(rand_v[patterni][1]-0.5));
   dradius[jid]=radius;
   dpoint[jid] = -1.0;
   dx=(v[0]-pp[jid][0]);
   dy=(v[1]-pp[jid][1]);
   dz=(v[2]-pp[jid][2]);
   if(fabs(dx) > radius)continue;
   if(fabs(dy) > radius)continue;
   if(fabs(dz) > radius)continue;
   if((a=sqrt(dx*dx+dy*dy+dz*dz)) < radius){
     dpoint[jid]=a;
     dc=(radius-a);
     if(dc > zb){
       id=jid;
       zb=dc;
     }
   }
 }
 if(id >= 0)
 {
     VECCOPY(pp[id],ppc)
     rad=dradius[id];
     VECSUB(v,ppc,u)          //(v-pp) VECTOR from centre of pattern
     if((a=sqrt((u[0])*(u[0])+
                (u[1])*(u[1])+
                (u[2])*(u[2]))) < rad)
     {
       float mu,moddn,rmu;
       VECTOR c,dn;
       mu=DOT(u,n);
       c[0]=ppc[0]+mu*n[0];
       c[1]=ppc[1]+mu*n[1];
       c[2]=ppc[2]+mu*n[2];
       VECSUB(v,c,dn)
       moddn=sqrt(dn[0]*dn[0]+dn[1]*dn[1]+dn[2]*dn[2]);
       rmu=sqrt(rad*rad-mu*mu);
       if(moddn < rmu)
       {
         a=1.0-moddn/rmu;
         normalize(dn);
         //
         // Quantize the direction of the "bending" vector so that
         // flat faces are generated as opposed to overlapping dimples.
         //
         dn[0]=0.5*(float)((int)(dn[0]*2.0));
         dn[1]=0.5*(float)((int)(dn[1]*2.0));
         dn[2]=0.5*(float)((int)(dn[2]*2.0));
         //
         normalize(dn);
         VECSCALE(-1.0,dn,dn);  // Invert for depressions.
         a=0.7;                 // Fixed slope.
         VECSUM((1.0-a)*dn,(a)*n,n)
         normalize(n);
       }
     }
 }
 //
 //  Next summation at higher frequency.
 //
 xx *= 2; yy *= 2; zz *=2;

 }  // end of k loop
 return TRUE;
}
